/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 * 
 */
package eu.semiramis.identityaggregator.client;

import eu.semiramis.commons.audit.AuditFactory;
import eu.semiramis.commons.audit.Transaction;
import javax.servlet.ServletRequest;

/**
 * Identity Aggregator - Audit Module
 * 
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class Audit {

    public interface EventId {

        public String getEventId();
    }

    public enum Level {

        DEBUG, ERROR, INFO, TRACE, WARN
    }

    public enum AppLogEventId implements EventId { // Event ids for application logging

        REQUEST_PROCESSING_INFO, REQUEST_PROCESSING_ERROR,
        RESPONSE_PROCESSING_INFO, RESPONSE_PROCESSING_ERROR, RESPONSE_STATUS_NOT_SUCCESS,
        REQUEST_GENERATION_ERROR, OUTGOING_REQUEST_INFO, OUTGOING_RESPONSE_INFO,
        SIGNATURE_VALIDATION_INFO, SIGNATURE_VALIDATION_ERROR, SIGNATURE_VALIDATED,
        AUTHENTICATION_TOKEN_PROCESSING_INFO, AUTHENTICATION_TOKEN_PROCESSING_ERROR, AUTHENTICATION_TOKEN_FOUND, AUTHENTICATION_TOKEN_NOT_FOUND,
        AUTHORIZATION_TOKEN_PROCESSING_INFO, AUTHORIZATION_TOKEN_PROCESSING_ERROR, AUTHORIZATION_TOKEN_FOUND, AUTHORIZATION_TOKEN_NOT_FOUND,
        ATTRIBUTE_PROCESSING_INFO, ATTRIBUTE_PROCESSING_ERROR, ATTRIBUTES_FOUND, NO_ATTRIBUTES_FOUND,
        ASSERTION_PROCESSING_INFO, ASSERTION_PROCESSING_ERROR,
        CONDITION_GENERATION_INFO, CONDITIONS_NOT_FULFILLED,
        NO_PROVIDER_INFO_FOUND,
        ORIGIN_ISSUER_PROCESSING_INFO, ORIGIN_ISSUER_PROCESSING_ERROR, ORIGIN_ISSUER_FOUND, ORIGIN_ISSUER_NOT_FOUND;

        @Override
        public String getEventId() {
            return this.name();
        }
    }
    public static final String TRANSACTION_ATTRIBUTE = "AuditTransaction";
    private static boolean initialized = false;
    private static ClientConfig clientConfig;

    public static void initAudit(ClientConfig config) {
        if (!initialized) {
            clientConfig = config;
            if (config.ENABLE_LOG_TO_AUDIT_SYSTEM) {
                AuditFactory.init(config.AUDIT_SERVER_CONFIG_PATH, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD, config.SEMIRAMIS_CLIENT_DOMAIN, config.SEMIRAMIS_CLIENT_ID);
            }
        }
    }
    private AuditDelegate appLogAuditDelegate;

    public Audit(String handlerName) {
        if (clientConfig.ENABLE_LOG_TO_AUDIT_SYSTEM) {
            this.appLogAuditDelegate = new AuditServer(handlerName);
        } else {
            this.appLogAuditDelegate = new AuditLocal(handlerName, clientConfig.LOG_CONFIG_PATH);
        }
    }

    public static void initRequest(ServletRequest request) {
        request.setAttribute(TRANSACTION_ATTRIBUTE, new Transaction());
    }

    public void audit(Level level, ServletRequest request, AppLogEventId eventId, String msg) {
        audit(level, getTransaction(request), eventId, msg);
    }

    public void audit(Level level, ServletRequest request, AppLogEventId eventId, Exception ex) {
        audit(level, getTransaction(request), eventId, ex);
    }

    public void audit(Level level, Transaction transaction, AppLogEventId eventId, String msg) {
        appLogAuditDelegate.audit(level, transaction, eventId, msg);
    }

    public void audit(Level level, Transaction transaction, AppLogEventId eventId, Exception ex) {
        appLogAuditDelegate.audit(level, transaction, eventId, ex);
    }

    private Transaction getTransaction(ServletRequest request) {
        Object transaction = request.getAttribute(TRANSACTION_ATTRIBUTE);
        if (transaction == null) {
            throw new IllegalStateException("Request not initialized for auditing");
        } else {
            return (Transaction) transaction;
        }
    }
}
