/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 * 
 */
package eu.semiramis.identityaggregator.client;

import eu.semiramis.commons.SAMLAttribute;
import eu.semiramis.identityaggregator.commons.Attribute;
import eu.semiramis.identityaggregator.commons.Conditions;
import eu.semiramis.identityaggregator.commons.AttributeToken;
import eu.semiramis.identityaggregator.commons.Token;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.client.Audit.Level;
import eu.semiramis.identityaggregator.client.TokenException.Type;
import eu.semiramis.identityaggregator.commons.Provider;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.InvalidPropertiesFormatException;
import java.util.Iterator;
import java.util.List;
import javax.servlet.http.*;

import org.opensaml.Configuration;
import org.opensaml.common.SAMLObjectBuilder;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.AuthnRequest;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;
import org.opensaml.xml.ConfigurationException;
import org.opensaml.xml.XMLObjectBuilderFactory;
import org.opensaml.xml.security.SecurityException;
import org.opensaml.common.SAMLObject;
import org.opensaml.common.SignableSAMLObject;
import org.opensaml.common.binding.decoding.SAMLMessageDecoder;
import org.opensaml.saml2.binding.decoding.HTTPPostDecoder;
import org.opensaml.saml2.binding.decoding.HTTPRedirectDeflateDecoder;
import org.opensaml.saml2.core.AttributeQuery;
import org.opensaml.saml2.core.Audience;
import org.opensaml.saml2.core.AudienceRestriction;
import org.opensaml.saml2.core.AuthzDecisionQuery;
import org.opensaml.saml2.core.EncryptedAssertion;
import org.opensaml.saml2.core.IDPEntry;
import org.opensaml.saml2.core.IDPList;
import org.opensaml.saml2.core.OneTimeUse;
import org.opensaml.saml2.core.RequestAbstractType;
import org.opensaml.saml2.core.Scoping;
import org.opensaml.saml2.core.StatusResponseType;
import org.opensaml.saml2.core.SubjectConfirmation;
import org.opensaml.saml2.core.SubjectConfirmationData;
import org.opensaml.xml.parse.XMLParserException;
import org.opensaml.xml.validation.ValidationException;

/**
 *  Client for performing request to Identity Aggregators
 * 
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IAClient {

    public static final String TRANSACTION_ATTRIBUTE = "AuditTransaction";
    static boolean initialized = false;
    private ClientConfig config = null;
    private Audit audit;

    /**
     * Constructor indicating the config file path
     * @param fileConfigPath config file path
     * @throws java.util.InvalidPropertiesFormatException
     * @throws java.io.FileNotFoundException
     * @throws java.io.IOException
     * @throws java.lang.NumberFormatException
     * @throws org.opensaml.xml.security.SecurityException
     * @throws java.lang.IllegalArgumentException
     * @throws java.lang.NoSuchFieldException
     * @throws java.lang.IllegalAccessException
     */
    public IAClient(String fileConfigPath) throws InvalidPropertiesFormatException, FileNotFoundException,
            IOException, NumberFormatException, SecurityException,
            IllegalArgumentException, NoSuchFieldException,
            IllegalAccessException,
            XMLParserException,
            ConfigurationException {
        config = new ClientConfig(fileConfigPath);
        if (!initialized) {
            SAMLUtil.startLibrary();
            Audit.initAudit(config);
            initialized = true;
        }
        audit = new Audit(IAClient.class.getName());

    }

    /**
     * Checks if the incoming request has been authenticated, that is, checks if contains a valid SAMLAuthenticationStatement
     * @param request HTTP Post request
     * @return true if has been successfully authenticated, false otherwise
     */
    public boolean containsAuthenticationToken(HttpServletRequest request) throws TokenException {
        return containsAuthenticationToken(request, getTransaction(request));
    }

    /**
     * Checks if the incoming request has been authenticated, that is, checks if contains a valid SAMLAuthenticationStatement
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return true if has been successfully authenticated, false otherwise
     */
    public boolean containsAuthenticationToken(HttpServletRequest request, Transaction auditTransaction) throws TokenException {
        // if is a SAMLResponse search Statements
        String samlMessage = request.getParameter("SAMLResponse");
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
        if (samlMessage != null) {
            //decode SAMLResponse
            Response samlResponse = null;
            try {
                samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
                throw new TokenException(Type.MalformedResponse);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, SAMLUtil.SAMLtoString(samlResponse));
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating Signature");
            try {
                validateSignature(samlResponse, auditTransaction);
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                throw new TokenException(Type.InvalidSignature);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Done.");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Searching Authentication Token...");
            return checkAuthentication(samlResponse, auditTransaction);
        } else {
            // if is a SAMLRequest search in extensions
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching SAMLRequest...");
            samlMessage = request.getParameter("SAMLRequest");
            if (samlMessage != null) {
                RequestAbstractType samlRequest = null;
                try {
                    samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, getDecoder(request, auditTransaction));
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
                    throw new TokenException(Type.MalformedResponse);
                }
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest found");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, SAMLUtil.SAMLtoString(samlRequest));
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating Signature");
                try {
                    validateSignature(samlRequest, auditTransaction);
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                    throw new TokenException(Type.InvalidSignature);
                }
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Done.");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching Tokens in extensions");
                List<Assertion> assertions = SAMLUtil.recoverAssertions(samlRequest.getExtensions());
                Iterator<Assertion> it = assertions.iterator();
                while (it.hasNext()) {
                    Assertion assertion = it.next();
                    if (!assertion.getAuthnStatements().isEmpty()) {
                        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_FOUND, "Authentication Token found");
                        return true;
                    }
                }
            }
        }
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_NOT_FOUND, "Authentication Token not found");
        return false;
    }

    /**
     * Checks if the incoming request contains attributes, that is, checks if contains SAMLAttributeStatements
     * @param request HTTP Post request
     * @return true if contains attributes, false otherwise
     */
    public boolean containsAttributes(HttpServletRequest request) throws TokenException {
        return containsAttributes(request, getTransaction(request));
    }

    /**
     * Checks if the incoming request contains attributes, that is, checks if contains SAMLAttributeStatements
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return true if contains attributes, false otherwise
     */
    public boolean containsAttributes(HttpServletRequest request, Transaction auditTransaction) throws TokenException {
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
        String samlMessage = request.getParameter("SAMLResponse");
        if (samlMessage != null) {
            Response samlResponse = null;
            try {
                samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
                throw new TokenException(Type.MalformedResponse);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, SAMLUtil.SAMLtoString(samlResponse));
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating signature...");
            try {
                validateSignature(samlResponse, auditTransaction);
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                throw new TokenException(Type.InvalidSignature);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Done.");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ATTRIBUTE_PROCESSING_INFO, "Searching Attributes in Response");
            return checkAttributes(samlResponse, auditTransaction);

        }
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.NO_ATTRIBUTES_FOUND, "Attributes not found");
        return false;
    }

    /**
     * Checks if the incoming request contains any authorization token, that is, checks if contains any SAMLAuthorizationStatement
     * @param request HTTP Post request
     * @return true if contains authorization tokens, false otherwise
     */
    public boolean containsAuthorizationToken(HttpServletRequest request) throws TokenException {
        return containsAuthorizationToken(request, getTransaction(request));
    }

    /**
     * Checks if the incoming request contains any authorization token, that is, checks if contains any SAMLAuthorizationStatement
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return true if contains authorization tokens, false otherwise
     */
    public boolean containsAuthorizationToken(HttpServletRequest request, Transaction auditTransaction) throws TokenException {
        // if is a SAMLResponse search Statements
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
        String samlMessage = request.getParameter("SAMLResponse");
        if (samlMessage != null) {
            //decode SAMLResponse
            Response samlResponse = null;
            try {
                samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
                throw new TokenException(Type.MalformedResponse);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, SAMLUtil.SAMLtoString(samlResponse));
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating signature...");
            try {
                validateSignature(samlResponse, auditTransaction);
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                throw new TokenException(Type.InvalidSignature);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Done.");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_INFO, "Searching Authorization Tokens...");
            return checkAuthorization(samlResponse, auditTransaction);
        } else {
            // if is a SAMLRequest search in extensions
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching SAMLRequest...");
            samlMessage = request.getParameter("SAMLRequest");
            if (samlMessage != null) {
                RequestAbstractType samlRequest = null;
                try {
                    samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, getDecoder(request, auditTransaction));
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
                    throw new TokenException(Type.MalformedResponse);
                }
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest found");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, SAMLUtil.SAMLtoString(samlRequest));
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating Signature...");
                try {
                    validateSignature(samlRequest, auditTransaction);
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                    throw new TokenException(Type.InvalidSignature);
                }
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Done.");
                List<Assertion> assertions = SAMLUtil.recoverAssertions(samlRequest.getExtensions());
                Iterator<Assertion> it = assertions.iterator();
                while (it.hasNext()) {
                    Assertion assertion = it.next();
                    if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_FOUND, "Authorization Token found");
                        return true;
                    }
                }
            }
        }
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_NOT_FOUND, "Authorization Token not found");
        return false;
    }

    private SAMLMessageDecoder getDecoder(HttpServletRequest request, Transaction auditTransaction) {
        SAMLMessageDecoder decoder = null;
        if (request.getMethod().equals("GET")) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Using Get Decoder");
            decoder = new HTTPRedirectDeflateDecoder();
        } else if (request.getMethod().equals("POST")) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Using Post decoder");
            decoder = new HTTPPostDecoder();
        }
        return decoder;
    }

    /**
     * Gets the authentication token contained in the incoming HTTP request.
     * @param request HTTP Post request
     * @return Authentication token or null if authentication token is not present.
     */
    public AuthenticationToken getAuthenticationToken(HttpServletRequest request) {
        return getAuthenticationToken(request, getTransaction(request));
    }

    /**
     * Gets the authentication token contained in the incoming HTTP request.
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return Authentication token or null if authentication token is not present.
     */
    public AuthenticationToken getAuthenticationToken(HttpServletRequest request, Transaction auditTransaction) {
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
        String samlMessage = request.getParameter("SAMLResponse");
        if (samlMessage != null) {
            try {
                Response samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Getting Authentication Token from response...");
                Assertion assertion = getAuthenticationAssertion(samlResponse, auditTransaction);
                AuthenticationToken token = new AuthenticationToken(assertion);
                audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_FOUND, "Authentication Token successfully got");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Authentication Token: " + SAMLUtil.SAMLtoString(assertion));
                return token;
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_ERROR, ex);
                return null;
            }
        } else {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching SAMLRequest...");
            samlMessage = request.getParameter("SAMLRequest");
            if (samlMessage != null) {
                try {
                    RequestAbstractType samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, getDecoder(request, auditTransaction));
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest found");
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching Token in extension...");
                    List<Assertion> assertions = SAMLUtil.recoverAssertions(samlRequest.getExtensions());
                    Iterator<Assertion> it = assertions.iterator();
                    while (it.hasNext()) {
                        Assertion assertion = it.next();
                        if (!assertion.getAuthnStatements().isEmpty()) {
                            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Getting Authentication Token from extension");
                            AuthenticationToken authenticationToken = new AuthenticationToken(assertion);
                            //SAMLUtil.releaseAssertion(assertion);
                            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_FOUND, "Authentication Token successfully got");
                            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Authentication Token: " + SAMLUtil.SAMLtoString(assertion));
                            return authenticationToken;
                        }
                    }
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_ERROR, ex);
                }

            }
        }
        return null;
    }

    /**
     * Gets the origin issuer of the request specified in the incoming HTTP request.
     * @param request HTTP Post request
     * @return Origin issuer of the request or null not present.
     */
    public String getOriginIssuer(HttpServletRequest request) {
        return getOriginIssuer(request, getTransaction(request));
    }

    /**
     * Gets the origin issuer of the request specified in the incoming HTTP request.
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return Origin issuer of the request or null not present.
     */
    public String getOriginIssuer(HttpServletRequest request, Transaction auditTransaction) {
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching SAMLRequest...");
        String samlMessage = request.getParameter("SAMLRequest");
        if (samlMessage != null) {
            try {
                RequestAbstractType samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, getDecoder(request, auditTransaction));
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest found");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ORIGIN_ISSUER_PROCESSING_INFO, "Searching origin issuer in extension...");
                String originIssuer = null;
                X509Certificate cert = SAMLUtil.recoverX509Certificate(samlRequest.getExtensions());
                if (cert != null) {
                    originIssuer = cert.getSubjectDN().getName();
                } else {
                    originIssuer = SAMLUtil.recoverOriginIssuer(samlRequest.getExtensions());
                }
                if (originIssuer != null) {
                    audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.ORIGIN_ISSUER_FOUND, "Origin issuer successfully got");
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ORIGIN_ISSUER_PROCESSING_INFO, "Origin issuer: " + originIssuer);
                    return originIssuer;
                } else {
                    audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.ORIGIN_ISSUER_NOT_FOUND, "Origin issuer not found");
                    return null;
                }
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.ORIGIN_ISSUER_PROCESSING_ERROR, ex);
                return null;
            }
        } else {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, "SAMLRequest not found");
            return null;
        }
    }

    /**
     * Gets attributes contained in the incoming HTTP request.
     * @param request HTTP Post request
     * @return List of attributes, or null if not attribute statements has been included.
     */
    public List<Attribute> getAttributes(HttpServletRequest request) {
        return getAttributes(request, getTransaction(request));
    }

    /**
     * Gets attributes contained in the incoming HTTP request.
     * @param request HTTP Post request
     * @param auditTransaction transaction to be used for auditing
     * @return List of attributes, or null if not attribute statements has been included.
     */
    public List<Attribute> getAttributes(HttpServletRequest request, Transaction auditTransaction) {
        List<Attribute> attributes = new ArrayList<Attribute>();
        try {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
            Response samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ATTRIBUTE_PROCESSING_INFO, "Getting Attributes from response...");
            List assertionList = getAssertions(samlResponse, auditTransaction);
            Iterator assertionIt = assertionList.iterator();
            while (assertionIt.hasNext()) {
                Assertion assertion = (Assertion) assertionIt.next();
                if (!assertion.getAttributeStatements().isEmpty()) {
                    AttributeToken attributeToken = new AttributeToken(assertion);
                    attributes.addAll(attributeToken.getAttributes());
                }


                /*
                Iterator it = assertion.getAttributeStatements().iterator();
                while (it.hasNext()) {
                AttributeStatement attributeStatement = (AttributeStatement) it.next();
                Iterator<org.opensaml.saml2.core.Attribute> it2 = attributeStatement.getAttributes().iterator();
                while (it2.hasNext()) {
                org.opensaml.saml2.core.Attribute attribute = it2.next();
                Attribute attr = null;
                
                if(attribute.getAttributeValues().get(0) instanceof XSAnyImpl){
                XSAnyImpl value = (XSAnyImpl) attribute.getAttributeValues().get(0);
                attr = new Attribute(attribute.getName(),value.getTextContent());
                Iterator<XMLObject> it3 = value.getOrderedChildren().iterator();
                while (it3.hasNext()) {
                XMLObject xMLObject = it3.next();
                System.out.println(xMLObject.getElementQName());
                }
                
                }
                else if(attribute.getAttributeValues().get(0) instanceof XSString){
                XSString value = (XSString) attribute.getAttributeValues().get(0);
                attr = new Attribute(attribute.getName(),value.getValue());
                }
                audit.audit(Level.INFO, "Attribute="+attr.getId()+", value="+attr.getValue());
                attributes.add(attr);
                }
                }
                 *
                 */
            }
            return attributes;
        } catch (Exception ex) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.ATTRIBUTE_PROCESSING_ERROR, ex);
            return null;
        }
    }

    /**
     * Gets authorization token contained in the incoming HTTP request.
     * @param request HTTP Post Request
     * @return Authorization Token or null if does not contain any Authorization statement.
     */
    public AuthorizationToken getAuthorizationToken(HttpServletRequest request) {
        return getAuthorizationToken(request, getTransaction(request));
    }

    /**
     * Gets authorization token contained in the incoming HTTP request.
     * @param request HTTP Post Request
     * @param auditTransaction transaction to be used for auditing
     * @return Authorization Token or null if does not contain any Authorization statement.
     */
    public AuthorizationToken getAuthorizationToken(HttpServletRequest request, Transaction auditTransaction) {
        String samlMessage = request.getParameter("SAMLResponse");
        if (samlMessage != null) {
            try {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
                Response samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_INFO, "Getting Authorization Token from response...");
                Assertion assertion = getAuthorizationAssertion(samlResponse, auditTransaction);
                AuthorizationToken token = new AuthorizationToken(assertion);
                audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_FOUND, "Authorization Token successfully got");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_INFO, "Authorization Token: " + SAMLUtil.SAMLtoString(assertion));
                return token;
            } catch (Exception ex) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_ERROR, ex);
                return null;
            }
        } else {
            samlMessage = request.getParameter("SAMLRequest");
            if (samlMessage != null) {
                try {
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Searching SAMLRequest...");
                    RequestAbstractType samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, getDecoder(request, auditTransaction));
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest found");
                    audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_INFO, "Searching Token in extension...");
                    List<Assertion> assertions = SAMLUtil.recoverAssertions(samlRequest.getExtensions());
                    Iterator<Assertion> it = assertions.iterator();
                    while (it.hasNext()) {
                        Assertion assertion = it.next();
                        if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                            AuthorizationToken authorizationToken = new AuthorizationToken(assertion);
                            //SAMLUtil.releaseAssertion(assertion);
                            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_FOUND, "Authorization Token successfully got");
                            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_PROCESSING_INFO, "Authorization Token: " + SAMLUtil.SAMLtoString(assertion));
                            return authorizationToken;
                        }
                    }
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_PROCESSING_ERROR, ex);
                }

            }
        }
        return null;
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request
     * @param response
     * @return AuthenticationResponse indicating the result of the request
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response) {
        return requestAuthenticationToken(request, response, getTransaction(request));
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request
     * @param response
     * @param auditTransaction transaction to be used for auditing
     * @return AuthenticationResponse indicating the result of the request
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, Transaction auditTransaction) {
        return requestAuthenticationToken(request, response, getOriginIssuer(), null, auditTransaction);
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request
     * @param response
     * @param originIssuer Origin issuer of the request
     * @return AuthenticationResponse indicating the result of the request
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, String originIssuer) {
        return requestAuthenticationToken(request, response, originIssuer, getTransaction(request));
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request
     * @param response
     * @param originIssuer Origin issuer of the request
     * @param auditTransaction transaction to be used for auditing
     * @return AuthenticationResponse indicating the result of the request
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, String originIssuer, Transaction auditTransaction) {
        return requestAuthenticationToken(request, response, originIssuer, null, auditTransaction);
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param conditions Additional conditions which have to be taken into account when user is authenticated
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, Conditions conditions) {
        return requestAuthenticationToken(request, response, conditions, getTransaction(request));
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param conditions Additional conditions which have to be taken into account when user is authenticated
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, Conditions conditions, Transaction auditTransaction) {
        return requestAuthenticationToken(request, response, getOriginIssuer(), conditions, auditTransaction);
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param originIssuer Origin issuer of the request
     * @param conditions Additional conditions which have to be taken into account when user is authenticated
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, String originIssuer, Conditions conditions) {
        return requestAuthenticationToken(request, response, originIssuer, conditions, getTransaction(request));
    }

    /**
     * Authenticates an end user, redirecting her to the IA.
     * This methods creates a SAML Request, sending it to the IA.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param originIssuer Origin issuer of the request
     * @param conditions Additional conditions which have to be taken into account when user is authenticated
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthenticationToken(HttpServletRequest request, HttpServletResponse response, String originIssuer, Conditions conditions, Transaction auditTransaction) {
        try {
            if (config.RESPONSE_URL == null) {
                config.RESPONSE_URL = request.getRequestURL().toString();
            }
            Provider provider = getIAProvider();
            AuthnRequest authNRequest = SAMLUtil.createAuthNRequest(config.SEMIRAMIS_CLIENT_ID, originIssuer, provider.getUrl(), config.RESPONSE_URL);
            if (conditions != null) {
                insertConditions(authNRequest, conditions, auditTransaction);
            }

            if (provider.isSignMessages()) {
                SAMLUtil.signXML(authNRequest, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
            }
            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.OUTGOING_REQUEST_INFO, "outgoing SAML AuthnRequest: " + SAMLUtil.SAMLtoString(authNRequest));
            SAMLUtil.httpPost(response, provider.getUrl(), authNRequest, config.RESPONSE_URL, config.TEMPLATE);
            RequestResult requestResult = new RequestResult();
            requestResult.setId(authNRequest.getID());
            requestResult.setSAMLRequest(authNRequest);
            requestResult.setStatus(RequestResult.Status.REDIRECTED);
            return requestResult;
        } catch (Exception ex) {
            RequestResult requestResult = new RequestResult();
            requestResult.setStatus(RequestResult.Status.ERROR);
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_GENERATION_ERROR, ex);
            return requestResult;
        }
    }

    private Provider getIAProvider() {
        return config.Providers.get(config.SEMIRAMIS_IA_ID);
    }

    /**
     * Returns the origin issuer as specified in the configuration file. If not specified, it returns null
     * @return 
     */
    private String getOriginIssuer() {
        return config.SEMIRAMIS_ORIGIN_ISSUER;
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, String subjectId, List<Attribute> attributes) {
        return requestAttributeAssertion(request, response, subjectId, attributes, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, String subjectId, List<Attribute> attributes, File attrEncrX509Cert) {
        return requestAttributeAssertion(request, response, subjectId, attributes, attrEncrX509Cert, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, String subjectId, List<Attribute> attributes, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, attributes, subjectId, null, null, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, String subjectId, List<Attribute> attributes, File attrEncrX509Cert, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, attributes, subjectId, null, attrEncrX509Cert, null, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, File attrEncrX509Cert, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, attrEncrX509Cert, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, List<Token> tokens, Transaction auditTransaction) {
        String subjectId = getSubjectId(tokens.get(0));
        return requestAttributeAssertion(request, response, attributes, subjectId, null, tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, File attrEncrX509Cert, List<Token> tokens, Transaction auditTransaction) {
        String subjectId = getSubjectId(tokens.get(0));
        return requestAttributeAssertion(request, response, attributes, subjectId, null, attrEncrX509Cert, tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, List<Token> tokens, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, getOriginIssuer(), tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param originIssuer Origin issuer of the request.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, String originIssuer, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, originIssuer, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, File attrEncrX509Cert, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, attrEncrX509Cert, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param originIssuer Origin issuer of the request.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, String originIssuer, List<Token> tokens, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, originIssuer, null, tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, File attrEncrX509Cert, List<Token> tokens, Transaction auditTransaction) {
        try {
            return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, null, SAMLUtil.readX509Certificate(attrEncrX509Cert), tokens, auditTransaction);
        } catch (Exception ex) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_GENERATION_ERROR, ex);
            RequestResult requestResult = new RequestResult();
            requestResult.setStatus(RequestResult.Status.ERROR);
            return requestResult;
        }
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param originIssuer Origin issuer of the request.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, String originIssuer, X509Certificate attrEncrX509Cert, List<Token> tokens) {
        return requestAttributeAssertion(request, response, attributes, subjectId, subjectDomain, originIssuer, attrEncrX509Cert, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param originIssuer Origin issuer of the request.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(HttpServletRequest request, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, String originIssuer, X509Certificate attrEncrX509Cert, List<Token> tokens, Transaction auditTransaction) {
        return requestAttributeAssertion(request.getRequestURL().toString(), response, attributes, subjectId, subjectDomain, originIssuer, attrEncrX509Cert, tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param requestURL HttpServletRequest URL
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param originIssuer Origin issuer of the request.
     * @param attrEncrX509Cert X509 certificate to be used for attribute encryption.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAttributeAssertion(String requestURL, HttpServletResponse response, List<Attribute> attributes, String subjectId, String subjectDomain, String originIssuer, X509Certificate attrEncrX509Cert, List<Token> tokens, Transaction auditTransaction) {
        try {
            if (config.RESPONSE_URL == null) {
                config.RESPONSE_URL = requestURL;
            }
            Provider provider = getIAProvider();
            List<SAMLAttribute> samlAttributes = new ArrayList<SAMLAttribute>();
            for (Attribute attribute : attributes) {
                samlAttributes.add(attribute.toSAMLAttribute());
            }

            AttributeQuery attrQuery = SAMLUtil.createAttrQuery(subjectId, subjectDomain, config.SEMIRAMIS_CLIENT_ID, originIssuer, provider.getUrl(), samlAttributes, attrEncrX509Cert);

            if (tokens != null) {
                SAMLUtil.insertAssertions(attrQuery, getAssertions(tokens));
            }
            if (provider.isSignMessages()) {
                SAMLUtil.signXML(attrQuery, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
            }
            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.OUTGOING_REQUEST_INFO, "outgoing SAML AttributeQuery: " + SAMLUtil.SAMLtoString(attrQuery));
            SAMLUtil.httpPost(response, provider.getUrl(), attrQuery, config.RESPONSE_URL, config.TEMPLATE);
            RequestResult requestResult = new RequestResult();
            requestResult.setId(attrQuery.getID());
            requestResult.setSAMLRequest(attrQuery);
            requestResult.setStatus(RequestResult.Status.REDIRECTED);
            return requestResult;
        } catch (Exception ex) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_GENERATION_ERROR, ex);
            RequestResult requestResult = new RequestResult();
            requestResult.setStatus(RequestResult.Status.ERROR);
            return requestResult;
        }
    }

    private List<Assertion> getAssertions(List<Token> tokens) {
        List<Assertion> assertions = new ArrayList();
        if (tokens != null) {
            Iterator<Token> it = tokens.iterator();
            while (it.hasNext()) {
                Token token = it.next();
                assertions.add(token.getAssertion());
            }
        }
        return assertions;
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param action action which will be performed
     * @param resource resource which action will be performed.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String action, String resource) {
        return requestAuthorizationToken(request, response, subjectId, action, resource, getTransaction(request));
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param action action which will be performed
     * @param resource resource which action will be performed.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String action, String resource, Transaction auditTransaction) {
        return requestAuthorizationToken(request, response, subjectId, null, action, resource, null, auditTransaction);
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse         
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String action, String resource, List<Token> tokens) {
        return requestAuthorizationToken(request, response, action, resource, tokens, getTransaction(request));
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse         
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String action, String resource, List<Token> tokens, Transaction auditTransaction) {
        String subjectId = getSubjectId(tokens.get(0));
        return requestAuthorizationToken(request, response, subjectId, null, action, resource, tokens, auditTransaction);
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param subjectDomain Domain where this request will be redirected.
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String subjectDomain, String action, String resource, List<Token> tokens) {
        return requestAuthorizationToken(request, response, subjectId, subjectDomain, action, resource, tokens, getTransaction(request));
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param subjectDomain Domain where this request will be redirected.
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String subjectDomain, String action, String resource, List<Token> tokens, Transaction auditTransaction) {
        return requestAuthorizationToken(request, response, subjectId, subjectDomain, action, resource, getOriginIssuer(), tokens, auditTransaction);
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param subjectDomain Domain where this request will be redirected.
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param originIssuer Origin issuer of the request.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String subjectDomain, String action, String resource, String originIssuer, List<Token> tokens) {
        return requestAuthorizationToken(request, response, subjectId, subjectDomain, action, resource, originIssuer, tokens, getTransaction(request));
    }

    /**
     * Request authorization for a subject to perform an action over a resource.
     * This method send a SAMLAuthorizationQuery to the Identity Aggregator.
     * @param request HTTPServletRequest
     * @param response HTTPServletResponse
     * @param subjectId subject which will perform the action.
     * @param subjectDomain Domain where this request will be redirected.
     * @param action action which will be performed
     * @param resource resource on which action will be performed.
     * @param originIssuer Origin issuer of the request.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     */
    public RequestResult requestAuthorizationToken(HttpServletRequest request, HttpServletResponse response, String subjectId, String subjectDomain, String action, String resource, String originIssuer, List<Token> tokens, Transaction auditTransaction) {
        try {
            if (config.RESPONSE_URL == null) {
                config.RESPONSE_URL = request.getRequestURL().toString();
            }
            Provider provider = getIAProvider();
            AuthzDecisionQuery authzQuery = SAMLUtil.createAuthzQuery(subjectId, subjectDomain, action, resource, config.SEMIRAMIS_CLIENT_ID, originIssuer, provider.getUrl(), null);

            if (tokens != null) {
                SAMLUtil.insertAssertions(authzQuery, getAssertions(tokens));
            }
            if (provider.isSignMessages()) {
                SAMLUtil.signXML(authzQuery, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
            }
            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.OUTGOING_REQUEST_INFO, "outcoming SAML AuthorizationQuery: " + SAMLUtil.SAMLtoString(authzQuery));
            SAMLUtil.httpPost(response, provider.getUrl(), authzQuery, config.RESPONSE_URL, config.TEMPLATE);
            RequestResult requestResult = new RequestResult();
            requestResult.setId(authzQuery.getID());
            requestResult.setSAMLRequest(authzQuery);
            requestResult.setStatus(RequestResult.Status.REDIRECTED);
            return requestResult;
        } catch (Exception ex) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.REQUEST_GENERATION_ERROR, ex);
            RequestResult requestResult = new RequestResult();
            requestResult.setStatus(RequestResult.Status.ERROR);
            return requestResult;
        }
    }

    /**
     * checks if the incoming request contains a SAMLResponse
     * @param request
     * @return
     */
    public boolean containsResponse(HttpServletRequest request) {
        return (request.getParameter("SAMLResponse") != null);
    }

    /**
     * Get authentication and authorization tokens or attributes contained in the incoming request
     * @param request
     * @return ResponseStruct contained
     */
    public ResponseStruct processResponse(HttpServletRequest request) throws TokenException {
        return processResponse(request, getTransaction(request));
    }

    /**
     * Get authentication and authorization tokens or attributes contained in the incoming request
     * @param request
     * @param auditTransaction transaction to be used for auditing
     * @return ResponseStruct contained
     */
    public ResponseStruct processResponse(HttpServletRequest request, Transaction auditTransaction) throws TokenException {
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Searching SAMLResponse...");
        Response samlResponse = null;
        try {
            samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
        } catch (Exception ex) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
            throw new TokenException(Type.MalformedResponse);
        }
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse found");
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Generating response structure...");
        ResponseStruct responseStruct = new ResponseStruct();
        if (samlResponse.getStatus().getStatusCode().getValue().equals(StatusCode.SUCCESS_URI)) {
            //responseStruct.setStatus(VALID);
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating Signature...");
            try {
                validateSignature(samlResponse, auditTransaction);
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                throw new TokenException(Type.InvalidSignature);
            }
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "Signature valid.");
            // Recover the request which this response references
            String inResponseTo = SAMLUtil.searchInResponseTo(samlResponse);
            responseStruct.setInResponseTo(inResponseTo);
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "Getting tokens...");
            List<Token> tokens = new ArrayList<Token>();
            responseStruct.setTokens(tokens);
            List<Assertion> assertions = getAssertions(samlResponse, auditTransaction);
            Iterator<Assertion> it = assertions.iterator();
            try {
                while (it.hasNext()) {
                    Assertion assertion = it.next();
                    if (!assertion.getAuthnStatements().isEmpty()) {
                        AuthenticationToken authToken = new AuthenticationToken(assertion);
                        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHENTICATION_TOKEN_FOUND, "Authentication Token found");
                        tokens.add(authToken);
                    }
                    if (!assertion.getAttributeStatements().isEmpty()) {
                        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ATTRIBUTES_FOUND, "Attribute Token found");
                        AttributeToken attributeToken = new AttributeToken(assertion);
                        tokens.add(attributeToken);
                    }
                    if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.AUTHORIZATION_TOKEN_FOUND, "Authorization Token found");
                        AuthorizationToken authorizationToken = new AuthorizationToken(assertion);
                        tokens.add(authorizationToken);
                    }
                }
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
                throw new TokenException(Type.MalformedResponse);
            }
        }
        return responseStruct;
    }


    /*
    public void postSAMLResponse(List<Token> tokens, String destination, String issuer){
    throw new UnsupportedOperationException("Not yet implemented");
    }
     *
     */
    /**
     * Gets the subject pseudonym through a token
     * @param token Token containing information related with the user
     * @return pseudonym of the user
     */
    public String getSubjectId(Token token) {
        return token.getAssertion().getSubject().getNameID().getValue();
    }

    /**
     * Checks if a token is valid. That is, it is propperly signed and it is still valid.
     * It also checks if the token has been issued by a relying party.
     * @param token Token which will be validated.
     * @return true if is still valid, false otherwise.
     */
    public boolean validateToken(Token token) {
        return validateToken(token, new Transaction());
    }

    /**
     * Checks if a token is valid. That is, it is propperly signed and it is still valid.
     * It also checks if the token has been issued by a relying party.
     * @param request HttpServletRequest
     * @param token Token which will be validated.
     * @return true if is still valid, false otherwise.
     */
    public boolean validateToken(HttpServletRequest request, Token token) {
        return validateToken(token, getTransaction(request));
    }

    /**
     * Checks if a token is valid. That is, it is propperly signed and it is still valid.
     * It also checks if the token has been issued by a relying party.
     * @param token Token which will be validated.
     * @param auditTransaction transaction to be used for auditing
     * @return true if is still valid, false otherwise.
     */
    public boolean validateToken(Token token, Transaction auditTransaction) {
        String issuer = token.getAssertion().getIssuer().getValue();
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Searching certificate for: " + issuer);
        if (!config.Providers.containsKey(issuer)) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.NO_PROVIDER_INFO_FOUND, "No Provider info found for Issuer \"" + issuer + "\"");
            return false;
        }
        Provider provider = config.Providers.get(issuer);
        String certPath = provider.getCertPath();
        /**@TODO Get certificates from external provider e.g. using XKMS */
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating XML signature");
        try {
            SAMLUtil.validateSign(token.getAssertion(), certPath);
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATED, "XML signature valid");
            return true;
        } catch (Exception ex) {
            audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
            return false;
        }
    }

    /**
     * Having a pseudonym, this method asks for real or domain specific identifier of a subject.
     * @param id subject id.
     * @return The real or domain specific identifier of the user.
     */
    public String resolveIDMapping(String id) {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    private List<Assertion> getAssertions(Response samlResponse, Transaction auditTransaction) {
        List<Assertion> assertions = new ArrayList<Assertion>();
        if (samlResponse.getAssertions() != null && !samlResponse.getAssertions().isEmpty()) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, "Assertions found");
            assertions.addAll(samlResponse.getAssertions());
        }
        if (samlResponse.getEncryptedAssertions() != null && !samlResponse.getEncryptedAssertions().isEmpty()) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, "Encrypted Assertions found");
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, "Decrypting assertions");
            List<Assertion> decryptedAssertions = decryptAssertions(samlResponse.getEncryptedAssertions(), auditTransaction);
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, "Assertions decrypted");
            assertions.addAll(decryptedAssertions);
        }
        return assertions;
    }

    private List<Assertion> decryptAssertions(List<EncryptedAssertion> encryptedAssertions, Transaction auditTransaction) {
        List<Assertion> assertions = new ArrayList<Assertion>();
        Iterator<EncryptedAssertion> it = encryptedAssertions.iterator();
        while (it.hasNext()) {
            EncryptedAssertion encryptedAssertion = it.next();
            try {
                assertions.add(SAMLUtil.decryptAssertion(encryptedAssertion, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD));
            } catch (Exception ex) {
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_ERROR, "Cannot decrypt assertion");
                audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_ERROR, ex);
            }
        }
        return assertions;

    }

    /**
     * checks if SAML response contains a valid Authentication Statement
     * @param samlResponse
     * @param auditTransaction transaction to be used for auditing
     * @return true if contains a valid SAMLAuthenticationStatmente, false otherwise
     */
    private boolean checkAuthentication(Response samlResponse, Transaction auditTransaction) throws TokenException {
        String status = samlResponse.getStatus().getStatusCode().getValue();
        if (!status.equals(StatusCode.SUCCESS_URI)) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_STATUS_NOT_SUCCESS, "SAMLResponse status not success");
            throw new TokenException(status, samlResponse.getStatus().getStatusMessage().getMessage());
        }

        boolean containAuthenticationAssertion = false;

        List assertionList = getAssertions(samlResponse, auditTransaction);
        Iterator assertionIt = assertionList.iterator();
        while (assertionIt.hasNext()) {
            Assertion assertion = (Assertion) assertionIt.next();
            if (assertion.getAuthnStatements() != null && !assertion.getAuthnStatements().isEmpty()) {
                containAuthenticationAssertion = true;
                if (checkConditions(assertion, samlResponse.getIssuer().getValue(), auditTransaction)) {
                    return true;
                }
            }
        }

        if (containAuthenticationAssertion) {
            throw new TokenException(Type.InvalidConditions);
        }
        return false;
    }

    /**
     * checks if SAML response contains a valid Authentication Statement
     * @param samlResponse
     * @param auditTransaction transaction to be used for auditing
     * @return true if contains a valid SAMLAuthenticationStatmente, false otherwise
     */
    private boolean checkAuthorization(Response samlResponse, Transaction auditTransaction) throws TokenException {
        String status = samlResponse.getStatus().getStatusCode().getValue();
        if (!status.equals(StatusCode.SUCCESS_URI)) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_STATUS_NOT_SUCCESS, "SAMLResponse status not success");
            throw new TokenException(status, samlResponse.getStatus().getStatusMessage().getMessage());
        }

        boolean containAuthenticationAssertion = false;

        List assertionList = getAssertions(samlResponse, auditTransaction);
        Iterator assertionIt = assertionList.iterator();
        while (assertionIt.hasNext()) {
            Assertion assertion = (Assertion) assertionIt.next();
            if (assertion.getAuthzDecisionStatements() != null && !assertion.getAuthzDecisionStatements().isEmpty()) {
                containAuthenticationAssertion = true;
                if (checkConditions(assertion, samlResponse.getIssuer().getValue(), auditTransaction)) {
                    return true;
                }
            }
        }

        if (containAuthenticationAssertion) {
            throw new TokenException(Type.InvalidConditions);
        }
        return false;
    }

    /**
     * Checks if conditions of an assertion are valid
     * @param subjectConfirmationData
     * @param auditTransaction transaction to be used for auditing
     * @return true if valid, false otherwise
     */
    private boolean checkConditions(Assertion assertion, String issuer, Transaction auditTransaction) {
        Provider provider = config.Providers.get(issuer);
        if (!provider.isCheckAssertionTime()) {
            return true;
        }

        if ((assertion.getConditions() != null) && (assertion.getConditions().getNotOnOrAfter() != null)) {
            if (assertion.getConditions().getNotOnOrAfter().isBeforeNow()) {
                audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.CONDITIONS_NOT_FULFILLED, "Assertion condition not fulfilled: NotOnOrAfter=" + assertion.getConditions().getNotOnOrAfter());
                return false;
            }
        }

        if (!assertion.getSubject().getSubjectConfirmations().isEmpty()) {
            Iterator subjectConfirmationIt = assertion.getSubject().getSubjectConfirmations().iterator();
            while (subjectConfirmationIt.hasNext()) {
                SubjectConfirmation subjectConfirmation = (SubjectConfirmation) subjectConfirmationIt.next();
                SubjectConfirmationData subjectConfirmationData = subjectConfirmation.getSubjectConfirmationData();
                if (!subjectConfirmationData.getNotOnOrAfter().isBeforeNow()) {
                    return true;
                } else {
                    audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.CONDITIONS_NOT_FULFILLED, "Assertion condition not fulfilled: NotOnOrAfter=" + subjectConfirmationData.getNotOnOrAfter());
                }
            }
            return false;
        }

        return true;
    }

    /**
     * checks if SAML Response contains Attribute Statements
     * @param samlResponse
     * @param auditTransaction transaction to be used for auditing
     * @return
     */
    private boolean checkAttributes(Response samlResponse, Transaction auditTransaction) throws TokenException {
        String status = samlResponse.getStatus().getStatusCode().getValue();
        if (!status.equals(StatusCode.SUCCESS_URI)) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.RESPONSE_STATUS_NOT_SUCCESS, "SAMLResponse status not success");
            throw new TokenException(status, samlResponse.getStatus().getStatusMessage().getMessage());
        }

        boolean containAuthenticationAssertion = false;

        List assertionList = getAssertions(samlResponse, auditTransaction);
        Iterator assertionIt = assertionList.iterator();
        while (assertionIt.hasNext()) {
            Assertion assertion = (Assertion) assertionIt.next();
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, SAMLUtil.SAMLtoString(assertion));
            if (assertion.getAttributeStatements() != null && !assertion.getAttributeStatements().isEmpty()) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ATTRIBUTES_FOUND, "Attribute Statements found");
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.ASSERTION_PROCESSING_INFO, SAMLUtil.SAMLtoString(assertion));
                containAuthenticationAssertion = true;
                if (checkConditions(assertion, samlResponse.getIssuer().getValue(), auditTransaction)) {
                    return true;
                }
            }
        }
        if (containAuthenticationAssertion) {
            throw new TokenException(Type.InvalidConditions);
        }
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.NO_ATTRIBUTES_FOUND, "Attributes not found");
        return false;
    }

    private Assertion getAuthenticationAssertion(Response samlResponse, Transaction auditTransaction) {
        List assertionList = getAssertions(samlResponse, auditTransaction);
        Iterator assertionIt = assertionList.iterator();
        while (assertionIt.hasNext()) {
            Assertion assertion = (Assertion) assertionIt.next();
            if (!assertion.getAuthnStatements().isEmpty()) {
                return assertion;
            }
        }
        return null;
    }

    private Assertion getAuthorizationAssertion(Response samlResponse, Transaction auditTransaction) {
        List assertionList = getAssertions(samlResponse, auditTransaction);
        Iterator assertionIt = assertionList.iterator();
        while (assertionIt.hasNext()) {
            Assertion assertion = (Assertion) assertionIt.next();
            if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                return assertion;
            }
        }
        return null;
    }

    private void insertConditions(AuthnRequest authnRequest, Conditions conditions, Transaction auditTransaction) {
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting conditions...");
        XMLObjectBuilderFactory builderFactory = Configuration.getBuilderFactory();
        SAMLObjectBuilder builder;
        if (conditions.getTargetDomain() != null) {
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting target domain: " + conditions.getTargetDomain());
            // Set the domain which will really process the authnRequest
            builder = (SAMLObjectBuilder) builderFactory.getBuilder(Scoping.DEFAULT_ELEMENT_NAME);
            Scoping scoping = (Scoping) builder.buildObject();
            builder = (SAMLObjectBuilder) builderFactory.getBuilder(IDPList.DEFAULT_ELEMENT_NAME);
            IDPList idpList = (IDPList) builder.buildObject();
            builder = (SAMLObjectBuilder) builderFactory.getBuilder(IDPEntry.DEFAULT_ELEMENT_NAME);
            IDPEntry idpEntry = (IDPEntry) builder.buildObject();

            idpEntry.setProviderID(conditions.getTargetDomain());
            authnRequest.setScoping(scoping);
            scoping.setIDPList(idpList);
            idpList.getIDPEntrys().add(idpEntry);
        }

        if (conditions.getOneTimeUse() != null || conditions.getOneTimeUse() != null
                || conditions.getNotOnOrAfter() != null || conditions.getAudiencieRestriction() != null) {
            builder = (SAMLObjectBuilder) builderFactory.getBuilder(org.opensaml.saml2.core.Conditions.DEFAULT_ELEMENT_NAME);
            org.opensaml.saml2.core.Conditions samlConditions = (org.opensaml.saml2.core.Conditions) builder.buildObject();
            authnRequest.setConditions(samlConditions);

            if (conditions.getOneTimeUse() != null) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting OneTimeUse");
                builder = (SAMLObjectBuilder) builderFactory.getBuilder(OneTimeUse.DEFAULT_ELEMENT_NAME);
                OneTimeUse oneTimeUse = (OneTimeUse) builder.buildObject();
                samlConditions.getConditions().add(oneTimeUse);
            }
            if (conditions.getNotBefore() != null) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting NotBefore: " + conditions.getNotBefore());
                samlConditions.setNotBefore(conditions.getNotBefore());
            }
            if (conditions.getNotOnOrAfter() != null) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting NotOnOrAfter: " + conditions.getNotOnOrAfter());
                samlConditions.setNotOnOrAfter(conditions.getNotOnOrAfter());
            }
            if (conditions.getAudiencieRestriction() != null) {
                audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.CONDITION_GENERATION_INFO, "Inserting AudienceRestriction: " + conditions.getAudiencieRestriction());
                Iterator<String> it = conditions.getAudiencieRestriction().iterator();
                while (it.hasNext()) {
                    String audienceId = it.next();
                    builder = (SAMLObjectBuilder) builderFactory.getBuilder(AudienceRestriction.DEFAULT_ELEMENT_NAME);
                    AudienceRestriction audienceRestriction = (AudienceRestriction) builder.buildObject();
                    builder = (SAMLObjectBuilder) builderFactory.getBuilder(Audience.DEFAULT_ELEMENT_NAME);
                    Audience audience = (Audience) builder.buildObject();
                    audience.setAudienceURI(audienceId);
                    audienceRestriction.getAudiences().add(audience);
                    samlConditions.getAudienceRestrictions().add(audienceRestriction);
                }
            }
        }
    }

    /**
     * Gets the URL of the Identity Aggregator indicating where the messages will be sent.
     * @return
     */
    public String getIaUrl() {
        return config.Providers.get(config.SEMIRAMIS_IA_ID).getUrl();
    }

    public void setIaId(String iaId) {
        config.SEMIRAMIS_IA_ID = iaId;
    }

    public String getIaId() {
        return config.SEMIRAMIS_IA_ID;
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, response, providerId, providerCSUrl, new Transaction());
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletRequest request, HttpServletResponse response, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, response, providerId, providerCSUrl, getTransaction(request));
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String providerId, String providerCSUrl, Transaction auditTransaction) throws Exception {
        return sendResponse(requestId, tokens, response, StatusCode.SUCCESS_URI, providerId, providerCSUrl, auditTransaction);
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String statusCode, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, response, statusCode, null, providerId, providerCSUrl);
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String statusCode, String statusMessage, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, response, statusCode, statusMessage, providerId, providerCSUrl, new Transaction());
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletRequest request, HttpServletResponse response, String statusCode, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, request, response, statusCode, null, providerId, providerCSUrl);
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletRequest request, HttpServletResponse response, String statusCode, String statusMessage, String providerId, String providerCSUrl) throws Exception {
        return sendResponse(requestId, tokens, response, statusCode, statusMessage, providerId, providerCSUrl, getTransaction(request));
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String statusCode, String providerId, String providerCSUrl, Transaction auditTransaction) throws Exception {
        return sendResponse(requestId, tokens, response, statusCode, null, providerId, providerCSUrl, auditTransaction);
    }

    public Response sendResponse(String requestId, List<Token> tokens, HttpServletResponse response, String statusCode, String statusMessage, String providerId, String providerCSUrl, Transaction auditTransaction) throws Exception {
        List assertions = getAssertions(tokens);
        Provider provider = config.Providers.get(providerId);
        if (provider.isSignOutcomingTokens()) {
            Iterator<Assertion> it = assertions.iterator();
            while (it.hasNext()) {
                Assertion assertion = it.next();
                SAMLUtil.signXML(assertion, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
            }
        }
        Response outcomingResponse;
        if (provider.isEncryptAssertions()) {
            List<EncryptedAssertion> encryptedAssertions = new ArrayList<EncryptedAssertion>();
            Iterator<Assertion> it = assertions.iterator();
            while (it.hasNext()) {
                Assertion assertion = it.next();
                if (!config.Providers.containsKey(providerId)) {
                    audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.NO_PROVIDER_INFO_FOUND, "No Provider info found for Issuer \"" + providerId + "\"");
                }
                String certPath = provider.getCertPath();
                encryptedAssertions.add(SAMLUtil.encryptAssertion(assertion, certPath));
            }
            assertions = encryptedAssertions;
        }

        outcomingResponse = SAMLUtil.createResponse(config.SEMIRAMIS_CLIENT_ID, requestId, providerCSUrl, assertions);
        outcomingResponse.getStatus().getStatusCode().setValue(statusCode);
        if (statusMessage != null) {
            outcomingResponse.getStatus().getStatusMessage().setMessage(statusMessage);
        }

        if (provider.isSignMessages()) {
            SAMLUtil.signXML(outcomingResponse, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
        }
        audit.audit(Level.INFO, auditTransaction, Audit.AppLogEventId.OUTGOING_RESPONSE_INFO, "outgoing Response: " + SAMLUtil.SAMLtoString(outcomingResponse));
        SAMLUtil.httpPost(response, providerCSUrl, outcomingResponse, providerCSUrl, config.TEMPLATE);
        return outcomingResponse;
    }

    private void validateSignature(SignableSAMLObject samlMessage, Transaction auditTransaction) throws FileNotFoundException, CertificateException, ValidationException {
        String issuer = getIssuer(samlMessage);
        audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Searching certificate for: " + issuer);
        if (!config.Providers.containsKey(issuer)) {
            audit.audit(Level.ERROR, auditTransaction, Audit.AppLogEventId.NO_PROVIDER_INFO_FOUND, "No Provider info found for Issuer \"" + issuer + "\"");
            throw new CertificateException();
        }
        Provider provider = config.Providers.get(issuer);
        if (provider.isCheckSignatures()) {
            String certPath = provider.getCertPath();
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating XML signature");
            SAMLUtil.validateSign(samlMessage, certPath);
            audit.audit(Level.DEBUG, auditTransaction, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "XML signature valid");
        }
    }

    private String getIssuer(SAMLObject samlObject) {
        if (samlObject instanceof StatusResponseType) {
            return ((StatusResponseType) samlObject).getIssuer().getValue();
        }
        if (samlObject instanceof RequestAbstractType) {
            return ((RequestAbstractType) samlObject).getIssuer().getValue();
        }
        return null;
    }

    public String getResponseURL() {
        return config.RESPONSE_URL;
    }

    public void setResponseURL(String responseURL) {
        config.RESPONSE_URL = responseURL;
    }

    private Transaction getTransaction(HttpServletRequest request) {
        Transaction transaction;
        Object attribute = request.getAttribute(TRANSACTION_ATTRIBUTE);
        if (attribute != null) {
            try {
                transaction = (Transaction) attribute;
            } catch (ClassCastException ex) {
                transaction = new Transaction();
            }
        } else {
            transaction = new Transaction();
        }
        return transaction;
    }

    /**
     * Methods kept for backward compatibility with (id, value) only attributes.
     */
    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, String subjectId, List<String> attributeNames) {
        return requestAttributes(request, response, subjectId, attributeNames, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param subjectId Id of the user
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, String subjectId, List<String> attributeNames, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, subjectId, generateAttributes(attributeNames), auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, List<String> attributeNames, List<Token> tokens) {
        return requestAttributes(request, response, attributeNames, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, List<String> attributeNames, List<Token> tokens, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, generateAttributes(attributeNames), tokens, auditTransaction);
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, List<String> attributeNames, String subjectId, String subjectDomain, List<Token> tokens) {
        return requestAttributes(request, response, attributeNames, subjectId, subjectDomain, tokens, getTransaction(request));
    }

    /**
     * Request user attributes to the Identity Aggregator sending an SAMLAttributeQuery.
     * @param request HttpServletRequest
     * @param response HttpServletResponse
     * @param attributes Attributes which desire requesting. If null, a general SAMLAttributeQuery will be sent trying to recover all attributes associated to a user.
     * @param subjectId Id of the user.
     * @param subjectDomain Domain which will be recovered user attributes.
     * @param tokens List of tokens indicating the user has previously authenticated or authorized.
     * @param auditTransaction transaction to be used for auditing
     * @return RequestResult indicating the result of the request, and the ID of the request in order to associate the response with this request.
     * 
     * @deprecated Replaced by requestAttributeAssertion().
     */
    public RequestResult requestAttributes(HttpServletRequest request, HttpServletResponse response, List<String> attributeNames, String subjectId, String subjectDomain, List<Token> tokens, Transaction auditTransaction) {
        return requestAttributeAssertion(request, response, generateAttributes(attributeNames), subjectId, subjectDomain, tokens, auditTransaction);
    }

    private List<Attribute> generateAttributes(List<String> attributeNames) {
        List<Attribute> attributes = new ArrayList<Attribute>();
        for (String attributeName : attributeNames) {
            attributes.add(new Attribute(attributeName));
        }
        return attributes;
    }
}
