/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 * 
 */
package eu.semiramis.identityaggregator.client;

import org.opensaml.saml2.core.StatusCode;

/**
 * TokenException
 * 
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class TokenException extends Exception {

    public enum Type {

        InvalidSignature("Invalid signature"),
        InvalidConditions("Invalid conditions"),
        RequesterError("Requester error"),
        ResponderError("Responder error"),
        AuthorizationDenied("Authorization denied"),
        MalformedResponse("Malformed response"),
        AuthenticationFailed("Authentication failed"),
        InvalidAttributeName("Invalid attribute name"),
        InvalidNameIdPolicy("Invalid nameId policy"),
        NoAuthenticationContext("No authentication context"),
        NoAvailableIDP("No available IDP"),
        NoPassive("No passive"),
        NoSupportedIDP("No supported IDP"),
        PartialLogout("Partial logout"),
        ProxyCountExceeded("Proxy count exceeded"),
        RequestUnsupported("Request unsupported"),
        VersionMismatch("Version mismatch"),
        ResourceNotRecognized("Resource not recognized"),
        UnknownAttributeProfile("Unknown attribute profile"),
        UnknownPrincipal("Unknown principal"),
        UnsupportedBinding("Unsupported binding"),
        Unspecified("Unspecified error");
        private String description;

        Type(String description) {
            this.description = description;
        }

        public String getDescription() {
            return description;
        }
    };
    private Type type;

    TokenException(String SAMLStatus) {
        setSAMLStatus(SAMLStatus);
    }

    TokenException(String SAMLStatus, String msg) {
        super(msg);
        setSAMLStatus(SAMLStatus);
    }

    public TokenException() {
        type = Type.Unspecified;
    }

    public TokenException(Type type) {
        setType(type);
    }

    public TokenException(Type type, String msg) {
        super(msg);
        setType(type);
    }

    public Type getType() {
        return type;
    }

    public void setType(Type type) {
        this.type = type;
    }

    @Override
    public String getMessage() {

        String msg = super.getMessage();
        if (msg == null) {
            return this.type.getDescription();
        }
        return msg;
    }

    void setSAMLStatus(String SAMLStatus) {
        if (SAMLStatus.equals(StatusCode.REQUEST_DENIED_URI)) {
            setType(Type.AuthorizationDenied);
        } else if (SAMLStatus.equals(StatusCode.REQUESTER_URI)) {
            setType(Type.RequesterError);
        } else if (SAMLStatus.equals(StatusCode.RESPONDER_URI)) {
            setType(Type.ResponderError);
        } else if (SAMLStatus.equals(StatusCode.AUTHN_FAILED_URI)) {
            setType(Type.AuthenticationFailed);
        } else if (SAMLStatus.equals(StatusCode.INVALID_ATTR_NAME_VALUE_URI)) {
            setType(Type.InvalidAttributeName);
        } else if (SAMLStatus.equals(StatusCode.INVALID_NAMEID_POLICY_URI)) {
            setType(Type.InvalidNameIdPolicy);
        } else if (SAMLStatus.equals(StatusCode.NO_AUTHN_CONTEXT_URI)) {
            setType(Type.NoAuthenticationContext);
        } else if (SAMLStatus.equals(StatusCode.NO_AVAILABLE_IDP_URI)) {
            setType(Type.NoAvailableIDP);
        } else if (SAMLStatus.equals(StatusCode.NO_PASSIVE_URI)) {
            setType(Type.NoPassive);
        } else if (SAMLStatus.equals(StatusCode.NO_SUPPORTED_IDP_URI)) {
            setType(Type.NoSupportedIDP);
        } else if (SAMLStatus.equals(StatusCode.PARTIAL_LOGOUT_URI)) {
            setType(Type.PartialLogout);
        } else if (SAMLStatus.equals(StatusCode.PROXY_COUNT_EXCEEDED_URI)) {
            setType(Type.ProxyCountExceeded);
        } else if (SAMLStatus.equals(StatusCode.REQUEST_UNSUPPORTED_URI)) {
            setType(Type.RequestUnsupported);
        } else if (SAMLStatus.equals(StatusCode.REQUEST_VERSION_DEPRECATED_URI)) {
            setType(Type.VersionMismatch);
        } else if (SAMLStatus.equals(StatusCode.REQUEST_VERSION_TOO_HIGH_URI)) {
            setType(Type.VersionMismatch);
        } else if (SAMLStatus.equals(StatusCode.REQUEST_VERSION_TOO_LOW_URI)) {
            setType(Type.VersionMismatch);
        } else if (SAMLStatus.equals(StatusCode.RESOURCE_NOT_RECOGNIZED_URI)) {
            setType(Type.ResourceNotRecognized);
        } else if (SAMLStatus.equals(StatusCode.UNKNOWN_ATTR_PROFILE_URI)) {
            setType(Type.UnknownAttributeProfile);
        } else if (SAMLStatus.equals(StatusCode.UNKNOWN_PRINCIPAL_URI)) {
            setType(Type.UnknownPrincipal);
        } else if (SAMLStatus.equals(StatusCode.UNSUPPORTED_BINDING_URI)) {
            setType(Type.UnsupportedBinding);
        } else if (SAMLStatus.equals(StatusCode.VERSION_MISMATCH_URI)) {
            setType(Type.VersionMismatch);
        } else {
            type = Type.Unspecified;
        }
    }
}
