/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.client.IAClient;
import eu.semiramis.identityaggregator.client.RequestResult;
import eu.semiramis.identityaggregator.client.RequestResult.Status;
import eu.semiramis.identityaggregator.client.TokenException;
import eu.semiramis.identityaggregator.commons.Attribute;
import eu.semiramis.identityaggregator.commons.Token;
import java.io.IOException;
import java.security.MessageDigest;
import java.security.PublicKey;
import java.security.cert.CertificateException;
import java.security.cert.CertificateExpiredException;
import java.security.cert.CertificateNotYetValidException;
import java.security.cert.X509Certificate;
import java.security.interfaces.DSAPublicKey;
import java.security.interfaces.RSAPublicKey;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.crypto.interfaces.DHPublicKey;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.apache.commons.codec.binary.Base64;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;
import org.opensaml.xacml.ctx.DecisionType.DECISION;
import sun.security.x509.X500Name;

/**
 * AttributeRequestProcessor
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class AttributeRequestProcessor {

    protected IAClient iaClient;
    protected Audit audit;
    protected AuthzProvider authzProvider;
    protected ID_Discovery idDiscovery;
    protected AttributeTranslator attributeTranslator;
    protected ID_Translator idTranslator;
    protected IdentityAggregator_Config config;

    public AttributeRequestProcessor(IAClient iaClient, AuthzProvider authzProvider, ID_Discovery idDiscovery, AttributeTranslator attributeTranslator, ID_Translator idTranslator, IdentityAggregator_Config config) {
        this.iaClient = iaClient;
        this.audit = new Audit(AttributeRequestProcessor.class.getName());
        this.authzProvider = authzProvider;
        this.idDiscovery = idDiscovery;
        this.attributeTranslator = attributeTranslator;
        this.idTranslator = idTranslator;
        this.config = config;
    }

    protected void processAttributeRequest(RequestMessage requestMessage, HttpServletRequest request, HttpServletResponse response) throws IOException, TokenException, CertificateException, ServletException {
        Map<Destination, List<Attribute>> providersMap = idDiscovery.discoverAttrProviderURL(request, requestMessage);
        if (!providersMap.isEmpty()) {

            // Include subjectId and authorization tokens
            List<Token> tokens = RequestLinker.getAuthenticationTokens(requestMessage.getId());
            if (tokens != null) {
                RequestLinker.removeAuthenticationTokens(requestMessage.getId());
                requestMessage.getTokens().addAll(tokens);
                if (requestMessage.getSubjectId() == null || requestMessage.getSubjectId().isEmpty()) {
                    Iterator<Token> it = tokens.iterator();
                    if (it.hasNext()) {
                        String subjectId = it.next().getSubjectId();
                        requestMessage.changeSubjectId(subjectId);
                    }
                }
            }

            // Check policies
            Map<String, Decision> authzDecisions = authzProvider.isAttributeQueryAllowed(request, requestMessage.getIssuer(), getOriginIssuer(requestMessage), requestMessage.getSubjectId(), providersMap);
            List<String> unauthorizedAttributes = new ArrayList<String>();
            List<String> needConsentAttributes = new ArrayList<String>();
            List<String> approvedAttributes = new ArrayList<String>();
            boolean notification = false;
            for (Map.Entry<String, Decision> authzDecision : authzDecisions.entrySet()) {
                Decision decision = authzDecision.getValue();
                if (!decision.getDecision().equals(DECISION.Permit)) {
                    unauthorizedAttributes.add(authzDecision.getKey());
                }
                if (decision.getObligations().contains(Decision.Obligation.ASK_FOR_USER_CONSENT)) {
                    needConsentAttributes.add(authzDecision.getKey());
                } else {
                    approvedAttributes.add(authzDecision.getKey());
                }
                notification |= decision.getObligations().contains(Decision.Obligation.NOTIFY_USER);
            }
            if (!unauthorizedAttributes.isEmpty()) {
                audit.audit(Audit.Level.TRACE, request, Audit.AppLogEventId.UNAUTHORIZED_REQUEST, "Unauthorized request");
                String unauthorizedAttributesMsg = "";
                for (String unauthorizedAttribute : unauthorizedAttributes) {
                    unauthorizedAttributesMsg = unauthorizedAttributesMsg + unauthorizedAttribute + ", ";
                }
                Response samlResponse = sendErrorBack(requestMessage, request, response, StatusCode.REQUEST_DENIED_URI, "Attribute release not allowed by policies: " + unauthorizedAttributesMsg.substring(0, unauthorizedAttributesMsg.length() - 2));
                if (samlResponse != null) {
                    audit.audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                }
                return;
            }
            HttpSession session = request.getSession(true);
            session.setAttribute("NotificationNeeded", notification);
            if (!needConsentAttributes.isEmpty()) {
                session.setAttribute("RequestMessage", requestMessage);
                session.setAttribute("RequestURL", request.getRequestURL().toString());
                session.setAttribute("ProvidersMap", providersMap);
                askForUserConsent(request, response, providersMap, needConsentAttributes, approvedAttributes, getOriginIssuer(requestMessage), requestMessage.getAttrEncrX509Cert());
            } else {
                requestAttributes(requestMessage, response, providersMap, request.getRequestURL().toString(), audit.getRequestTransaction(request));
            }
        } else {
            idDiscovery.showWAYF(request, response, requestMessage);
        }
    }

    protected void requestAttributes(RequestMessage requestMessage, HttpServletResponse response, Map<Destination, List<Attribute>> providersMap, String requestURL, Transaction auditTransaction) throws CertificateException {
        Map<Destination, SubjectAttributesList> providersAttributesMap = translateProvidersMap(requestMessage, providersMap, auditTransaction);
        Destination destination = providersAttributesMap.keySet().iterator().next();
        List<Attribute> attributes = providersAttributesMap.get(destination).getAttributes();
        String subjectId = providersAttributesMap.get(destination).getSubjectId();
        iaClient.setIaId(destination.getProviderId());
        String originIssuer = requestMessage.getOriginIssuer();
        X509Certificate attrEncrX509Cert = requestMessage.getAttrEncrX509Cert();
        RequestResult requestResult = iaClient.requestAttributeAssertion(requestURL, response, attributes, subjectId, destination.getDomain(), originIssuer, attrEncrX509Cert, requestMessage.getTokens(), auditTransaction);
        if (requestResult.getStatus() == Status.REDIRECTED) {
            RequestLinker.linkRequest(requestResult.getId(), requestMessage);
            providersAttributesMap.remove(destination);
            RequestLinker.linkProvidersMap(requestMessage, providersAttributesMap);
            audit.audit_ATTR_REQ_SENT(Audit.Level.INFO, auditTransaction, requestResult.getId(), destination.getProviderId());
        } else if (requestResult.getStatus() == Status.ERROR) {
            Response samlResponse = sendErrorBack(requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error sending attribute request.", auditTransaction);
            if (samlResponse != null) {
                audit.audit_ATTR_RESP_SENT(Audit.Level.ERROR, auditTransaction, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

    private void askForUserConsent(HttpServletRequest request, HttpServletResponse response, Map<Destination, List<Attribute>> providersMap, List<String> needConsentAttributes, List<String> approvedAttributes, String attributeRequestor, X509Certificate attributeRequestorCert) throws ServletException, IOException {
        List<Attribute> needConsentAttrs = new ArrayList();
        List<Attribute> approvedAttrs = new ArrayList();
        for (Map.Entry<Destination, List<Attribute>> entry : providersMap.entrySet()) {
            for (Attribute attribute : entry.getValue()) {
                if (needConsentAttributes.contains(attribute.getId()) && !needConsentAttrs.contains(attribute)) {
                    needConsentAttrs.add(attribute);
                } else if (approvedAttributes.contains(attribute.getId()) && !approvedAttrs.contains(attribute)) {
                    approvedAttrs.add(attribute);
                }
            }
        }

        RequestorCertInfo requestorCertInfo = null;
        if (attributeRequestorCert != null) {
            requestorCertInfo = new RequestorCertInfo();
            try {
                requestorCertInfo.setVersion(attributeRequestorCert.getVersion());
                requestorCertInfo.setSerialNumber(toHexString(attributeRequestorCert.getSerialNumber().toByteArray()));
                requestorCertInfo.setIssuer(attributeRequestorCert.getIssuerDN().getName());
                requestorCertInfo.setSubject(attributeRequestorCert.getSubjectDN().getName());
                requestorCertInfo.setValidNotBefore(attributeRequestorCert.getNotBefore());
                requestorCertInfo.setValidNotAfter(attributeRequestorCert.getNotAfter());
                MessageDigest messagedigest = MessageDigest.getInstance("SHA");
                messagedigest.update(attributeRequestorCert.getEncoded());
                byte[] fingerprint = messagedigest.digest();
                requestorCertInfo.setFingerprint(toHexString(fingerprint));
                requestorCertInfo.setPublicKeyType(getPublicKeyType(attributeRequestorCert.getPublicKey()));
                requestorCertInfo.setPublicKey(toHexString(attributeRequestorCert.getPublicKey().getEncoded()));
                requestorCertInfo.setEncodedCertificate(new String(Base64.encodeBase64(attributeRequestorCert.getEncoded())));
                requestorCertInfo.setValid(checkCertificate(attributeRequestorCert));
            } catch (Exception e) {
                requestorCertInfo = null;
            }
        }
        request.setAttribute("Organization", config.ORGANIZATION);
        request.setAttribute("ContactInfo", config.CONTACT_INFO);
        request.setAttribute("Requestor", attributeRequestor);
        request.setAttribute("RequestorCertInfo", requestorCertInfo);
        request.setAttribute("NeedConsentAttributes", needConsentAttrs);
        request.setAttribute("ApprovedAttributes", approvedAttrs);

        request.getRequestDispatcher(config.USER_CONSENT_DISPATCHER).forward(request, response);
    }

    private Map<Destination, SubjectAttributesList> translateProvidersMap(RequestMessage requestMessage, Map<Destination, List<Attribute>> providersMap, Transaction transaction) {
        Map<Destination, SubjectAttributesList> providersAttributeMap = new HashMap<Destination, SubjectAttributesList>();
        Iterator<Destination> it = providersMap.keySet().iterator();
        while (it.hasNext()) {
            Destination destination = it.next();
            List<Attribute> attributes = providersMap.get(destination);
            String subjectId;
            if (destination.getDomain() != null) {
                attributes = translateRequestAttributes(requestMessage.getIssuer(), destination.getDomain(), attributes, requestMessage.getSubjectId(), transaction);
                subjectId = translateRequestSubjectId(requestMessage.getIssuer(), destination.getDomain(), requestMessage.getSubjectId(), transaction);
            } else {
                attributes = translateRequestAttributes(requestMessage.getIssuer(), destination.getProviderId(), attributes, requestMessage.getSubjectId(), transaction);
                subjectId = translateRequestSubjectId(requestMessage.getIssuer(), destination.getProviderId(), requestMessage.getSubjectId(), transaction);
            }
            SubjectAttributesList subjectAttributeList = new SubjectAttributesList();
            subjectAttributeList.setAttributes(attributes);
            subjectAttributeList.setSubjectId(subjectId);
            providersAttributeMap.put(destination, subjectAttributeList);
        }
        return providersAttributeMap;
    }

    protected List<Attribute> translateRequestAttributes(String issuer, String destination, List<Attribute> attributes, String subjectId, Transaction transaction) {
        return attributeTranslator.translateRequestAttributes(issuer, destination, attributes, subjectId, transaction);
    }

    private String translateRequestSubjectId(String issuer, String destination, String subjectId, Transaction transaction) {
        return idTranslator.translateRequestSubjectId(issuer, destination, subjectId, transaction);
    }

    protected Response sendErrorBack(RequestMessage requestMessage, HttpServletRequest request, HttpServletResponse response, String statusCode, String statusMessage) {
        try {
            return iaClient.sendResponse(requestMessage.getId(), null, request, response, statusCode, statusMessage, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
        } catch (Exception ex) {
            audit.audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            ex.printStackTrace();
            /**
             * @TODO show web error to the end user
             */
            return null;
        }
    }

    protected Response sendErrorBack(RequestMessage requestMessage, HttpServletResponse response, String statusCode, String statusMessage, Transaction transaction) {
        try {
            return iaClient.sendResponse(requestMessage.getId(), null, response, statusCode, statusMessage, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
        } catch (Exception ex) {
            audit.audit(Audit.Level.ERROR, transaction, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            ex.printStackTrace();
            /**
             * @TODO show web error to the end user
             */
            return null;
        }
    }

    protected String getOriginIssuer(RequestMessage requestMessage) {
        try {
            String originIssuer = null;
            X509Certificate cert = requestMessage.getAttrEncrX509Cert();
            if (cert != null) {
                X500Name dn = new X500Name(cert.getSubjectX500Principal().getName());
                originIssuer = dn.getCommonName();
            } else {
                originIssuer = requestMessage.getOriginIssuer();
            }
            return originIssuer;
        } catch (Exception ex) {
            ex.printStackTrace();
            return null;
        }
    }

    public String toHexString(byte[] bytes) {
        return toHexString(bytes, " ");
    }

    public String toHexString(byte[] bytes, String separator) {
        String string = "";
        for (byte b : bytes) {
            string = string + String.format("%02X", b) + separator;
        }
        return string.substring(0, string.length() - separator.length());
    }

    public String getPublicKeyType(PublicKey key) {
        if (key instanceof RSAPublicKey) {
            return key.getAlgorithm() + " Public Key (" + ((RSAPublicKey) key).getModulus().bitLength() + " bits)";
        } else if (key instanceof DSAPublicKey) {
            return key.getAlgorithm() + " Public Key (" + ((DSAPublicKey) key).getParams().getP().bitLength() + " bits)";
        } else if (key instanceof DHPublicKey) {
            return key.getAlgorithm() + " Public Key (" + ((DHPublicKey) key).getParams().getP().bitLength() + " bits)";
        } else {
            return "Unknown";
        }
    }

    public boolean checkCertificate(X509Certificate cert) {
        try {
// TODO - Validate signatures and check against trusted CAs
            cert.checkValidity();
            return true;
        } catch (CertificateExpiredException ex) {
            return false;
        } catch (CertificateNotYetValidException ex) {
            return false;
        }
    }
}
