/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.commons.Attribute;
import java.util.AbstractMap.SimpleEntry;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map.Entry;
import javax.servlet.http.HttpServletRequest;

/**
 * Identity Aggregator - Attribute Translator
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class AttributeTranslator {

    IdentityAggregator_Config config;
    Audit audit;

    public AttributeTranslator(IdentityAggregator_Config config) {
        this.config = config;
        audit = new Audit(AttributeTranslator.class.getName());
    }

    public List<Attribute> translateRequestAttributes(HttpServletRequest request, String issuer, String destination, List<Attribute> attributes, String subjectId) {
        return translateRequestAttributes(issuer, destination, attributes, subjectId, audit.getRequestTransaction(request));
    }

    public List<Attribute> translateRequestAttributes(String issuer, String destination, List<Attribute> attributes, String subjectId, Transaction transaction) {
        audit.audit(Audit.Level.INFO, transaction, Audit.AppLogEventId.ATTRIBUTE_TRANSLATION_INFO, "Translating attributes from " + issuer + " to " + destination);
        Entry<String, String> entry = new SimpleEntry<String, String>(issuer, destination);
        HashMap<String, String> attributeDictionary = config.AttributeDictionariesMap.get(entry);
        if (attributeDictionary == null) {
            audit.audit(Audit.Level.WARN, transaction, Audit.AppLogEventId.ATTRIBUTE_TRANSLATION_ERROR, "No dictionary configured");
            return attributes;
        }

        List<Attribute> translatedAttributes = new ArrayList<Attribute>();
        Iterator<Attribute> it = attributes.iterator();
        while (it.hasNext()) {
            Attribute attribute = it.next();
            Attribute translatedAttribute;
            String attributeId = attribute.getId();
            if (attributeDictionary.containsKey(attributeId)) {
                translatedAttribute = new Attribute(attributeDictionary.get(attributeId), attribute.getFriendlyName(), attribute.getExtendedInfo(), attribute.getXMLObjectValue());
            } else {
                translatedAttribute = attribute;
            }
            audit.audit_ATTR_TRANSLATION(Audit.Level.INFO, transaction, attribute.getId(), translatedAttribute.getId(), subjectId);
            translatedAttributes.add(translatedAttribute);
        }
        return translatedAttributes;
    }

    public List<Attribute> translateResponseAttributes(HttpServletRequest request, String destination, String issuer, List<Attribute> attributes, String subjectId) {
        return translateResponseAttributes(destination, issuer, attributes, subjectId, audit.getRequestTransaction(request));
    }

    public List<Attribute> translateResponseAttributes(String destination, String issuer, List<Attribute> attributes, String subjectId, Transaction transaction) {
        audit.audit(Audit.Level.INFO, transaction, Audit.AppLogEventId.ATTRIBUTE_TRANSLATION_INFO, "Translating attributes from " + destination + " to " + issuer);
        Entry<String, String> entry = new SimpleEntry<String, String>(destination, issuer);
        HashMap<String, String> attributeDictionary = config.AttributeDictionariesMap.get(entry);
        if (attributeDictionary == null) {
            audit.audit(Audit.Level.WARN, transaction, Audit.AppLogEventId.ATTRIBUTE_TRANSLATION_ERROR, "No dictionary configured");
            return attributes;
        }

        List<Attribute> translatedAttributes = new ArrayList<Attribute>();
        Iterator<Attribute> it = attributes.iterator();
        while (it.hasNext()) {
            Attribute attribute = it.next();
            Attribute translatedAttribute;
            if (attributeDictionary.containsKey(attribute.getId())) {
                translatedAttribute = new Attribute(attributeDictionary.get(attribute.getId()), attribute.getFriendlyName(), attribute.getExtendedInfo(), attribute.getXMLObjectValue());
            } else {
                translatedAttribute = attribute;
            }
            audit.audit_ATTR_TRANSLATION(Audit.Level.INFO, transaction, attribute.getId(), translatedAttribute.getId(), subjectId);
            translatedAttributes.add(translatedAttribute);
        }
        return translatedAttributes;
    }
}
