/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.AuditFactory;
import eu.semiramis.commons.audit.Transaction;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.servlet.ServletRequest;

/**
 * Identity Aggregator - Audit Module
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class Audit {

    public interface EventId {

        public String getEventId();
    }

    public enum Level {

        DEBUG, ERROR, INFO, TRACE, WARN
    }

    public enum SemiramisEventId implements EventId { // Event ids for SEMIRAMIS auditing

        AUTHZ_PERMITTED_BY_POLICY, // "Authorization granted by policy to (originIssuer) for attribute (attr name) of user (subjectIdentity)"
        AUTHZ_DENIED_BY_POLICY, // "Authorization denied by policy to (originIssuer) for attribute (attr name) of user (subjectIdentity)"

        AUTHZ_PERMITTED_BY_USER, // "Authorization granted by user consent to (originIssuer) for attributes (attr name 1), (attr name 2) ... (attr name n) of user (subjectIdentity)"
        AUTHZ_DENIED_BY_USER, // "Authorization denied by user consent to (originIssuer) for attributes (attr name 1), (attr name 2) ... (attr name n) of user (subjectIdentity)"

        ATTR_RELEASE, // "Attribute (attr name) of user (subjectIdentity) released to (original issuer)"
        // ATTR_RELEASE, // "Attributes (attr name 1), (attr name 2) ... (attr name n) of user (subjectIdentity) released to (original issuer)"
        AUTHZ_TOKEN_RELEASE, // "Authorization token released to (original issuer) for attributes (attr name 1), (attr name 2) ... (attr name n) of user (subjectIdentity) until (date)"
        ATTR_TRANSLATION, // "Attribute (attr name) of user (subjectIdentity) translated to (new attr name)"
        AUTHENTICATION, // "User (subjectIdentity) Authenticated to (original issuer)"

        AUTHN_REQ_SENT, // "Sent authentication request with id (messageID) to (destination)"
        AUTHN_RESP_SENT, // "Sent authentication response with id (messageID) to (destination)"
        AUTHN_REQ_RECEIVED, // "Received authentication request with id (messageID) from (destination)"
        AUTHN_RESP_RECEIVED, // "Received authentication response with id (messageID) from (destination)"

        ATTR_REQ_SENT, // "Sent attribute request with id (messageID) to (destination)"
        ATTR_RESP_SENT, // "Sent attribute response with id (messageID) to (destination)"
        ATTR_REQ_RECEIVED, // "Received attribute request with id (messageID) from (destination)"
        ATTR_RESP_RECEIVED, // "Received attribute response with id (messageID) from (destination)"

        AUTHZ_REQ_SENT, // "Sent Authorization request with id (messageID) to (destination)"
        AUTHZ_RESP_SENT, // "Sent Authorization response with id (messageID) to (destination)"
        AUTHZ_REQ_RECEIVED, // "Received Authorization request with id (messageID) from (destination)"
        AUTHZ_RESP_RECEIVED, // "Received Authorization response with id (messageID) from (destination)"

        UNKNOWN_MESSAGE_RECEIVED;  // "Unknown message received"

        @Override
        public String getEventId() {
            return this.name();
        }
    }

    public enum AppLogEventId implements EventId { // Event ids for application logging

        IA_INITIALIZATION_INFO, IA_INITIALIZATION_ERROR,
        IA_CS_INITIALIZATION_INFO, IA_CS_INITIALIZATION_ERROR,
        IA_CONNECTOR_INSTANTIATION_ERROR,
        TOKEN_PROCESSING_ERROR,
        SIGNATURE_VALIDATION_INFO, SIGNATURE_VALIDATION_ERROR, INVALID_SIGNATURE,
        REQUEST_PROCESSING_INFO, REQUEST_PROCESSING_ERROR,
        RESPONSE_PROCESSING_INFO, RESPONSE_PROCESSING_ERROR, RESPONSE_GENERATION_ERROR,
        ATTRIBUTE_TRANSLATION_INFO, ATTRIBUTE_TRANSLATION_ERROR,
        ID_TRANSLATION_INFO, ID_TRANSLATION_ERROR,
        AUTHORIZATION_INFO, AUTHORIZATION_ERROR, UNAUTHORIZED_REQUEST,
        POLICY_EVALUATION_INFO, POLICY_EVALUATION_ERROR, INTERNAL_POLICIES_NOT_FOUND,
        DISCOVERY_INFO, USER_ORIGIN_NOT_FOUND, DESTINATION_NOT_FOUND,
        WAYF_INFO, WAYF_ERROR;

        @Override
        public String getEventId() {
            return this.name();
        }
    }
    private static Map<ServletRequest, Transaction> transactions;
    public static final String TRANSACTION_ATTRIBUTE = "AuditTransaction";
    private static boolean initialized = false;
    private static IdentityAggregator_Config iaConfig;

    public static void initAudit(IdentityAggregator_Config config) {
        if (!initialized) {
            transactions = new HashMap<ServletRequest, Transaction>();
            iaConfig = config;
            if (config.ENABLE_AUDIT_SYSTEM) {
                AuditFactory.init(config.AUDIT_SERVER_CONFIG_PATH, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD, config.DOMAIN, config.ID);
            }
        }
    }
    private AuditDelegate semiramisAuditDelegate;
    private AuditDelegate appLogAuditDelegate;

    public Audit(String handlerName) {
        if (iaConfig.ENABLE_AUDIT_SYSTEM) {
            this.semiramisAuditDelegate = new AuditServer(handlerName);
            if (iaConfig.ENABLE_LOG_TO_AUDIT_SYSTEM) {
                this.appLogAuditDelegate = new AuditServer(handlerName);
            } else {
                this.appLogAuditDelegate = new AuditLocal(handlerName, iaConfig.LOG_CONFIG_PATH);
            }
        } else {
            this.semiramisAuditDelegate = new AuditLocal(handlerName, iaConfig.LOG_CONFIG_PATH);
            this.appLogAuditDelegate = new AuditLocal(handlerName, iaConfig.LOG_CONFIG_PATH);
        }
    }

    public static void initRequest(ServletRequest request) {
        transactions.put(request, new Transaction());
    }

    public static void clearRequest(ServletRequest request) {
        transactions.remove(request);
    }

    public void audit(Level level, ServletRequest request, SemiramisEventId eventId, String msg) {
        audit(level, getRequestTransaction(request), eventId, msg);
    }

    public void audit(Level level, ServletRequest request, SemiramisEventId eventId, Exception ex) {
        audit(level, getRequestTransaction(request), eventId, ex);
    }

    public void audit(Level level, Transaction transaction, SemiramisEventId eventId, String msg) {
        semiramisAuditDelegate.audit(level, transaction, eventId, msg);
    }

    public void audit(Level level, Transaction transaction, SemiramisEventId eventId, Exception ex) {
        semiramisAuditDelegate.audit(level, transaction, eventId, ex);
    }

    public void audit(Level level, ServletRequest request, AppLogEventId eventId, String msg) {
        audit(level, getRequestTransaction(request), eventId, msg);
    }

    public void audit(Level level, ServletRequest request, AppLogEventId eventId, Exception ex) {
        audit(level, getRequestTransaction(request), eventId, ex);
    }

    public void audit(Level level, Transaction transaction, AppLogEventId eventId, String msg) {
        appLogAuditDelegate.audit(level, transaction, eventId, msg);
    }

    public void audit(Level level, Transaction transaction, AppLogEventId eventId, Exception ex) {
        appLogAuditDelegate.audit(level, transaction, eventId, ex);
    }

    public Transaction getRequestTransaction(ServletRequest request) {
        Transaction transaction = transactions.get(request);
        if (transaction == null) {
//            throw new IllegalStateException("Request not initialized for auditing");
            return new Transaction();
        } else {
            return transaction;
        }
    }

    /*
     * Methods for SEMIRAMIS audit logging with fixed message format.
     */
    public void audit_AUTHZ_PERMITTED_BY_POLICY(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit_AUTHZ_PERMITTED_BY_POLICY(level, getRequestTransaction(request), originIssuer, attrNames, subjectIdentity);
    }

    public void audit_AUTHZ_DENIED_BY_POLICY(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit_AUTHZ_DENIED_BY_POLICY(level, getRequestTransaction(request), originIssuer, attrNames, subjectIdentity);
    }

//    public void audit_AUTHZ_PERMITTED_BY_USER(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity) {
//        audit_AUTHZ_PERMITTED_BY_USER(level, getTransaction(request), originIssuer, attrNames, subjectIdentity);
//    }
//    public void audit_AUTHZ_DENIED_BY_USER(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity) {
//        audit_AUTHZ_DENIED_BY_USER(level, getTransaction(request), originIssuer, attrNames, subjectIdentity);
//    }
//    public void audit_ATTR_RELEASE(Level level, ServletRequest request, String originIssuer, String attrName, String subjectIdentity) {
//        audit_ATTR_RELEASE(level, getTransaction(request), originIssuer, attrName, subjectIdentity);
//    }
    public void audit_ATTR_RELEASE(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit_ATTR_RELEASE(level, getRequestTransaction(request), originIssuer, attrNames, subjectIdentity);
    }

    public void audit_AUTHZ_TOKEN_RELEASE(Level level, ServletRequest request, String originIssuer, List<String> attrNames, String subjectIdentity, Date date) {
        audit_AUTHZ_TOKEN_RELEASE(level, getRequestTransaction(request), originIssuer, attrNames, subjectIdentity, date);
    }

    public void audit_ATTR_TRANSLATION(Level level, ServletRequest request, String attrName, String newAttrName, String subjectIdentity) {
        audit_ATTR_TRANSLATION(level, getRequestTransaction(request), attrName, newAttrName, subjectIdentity);
    }

    public void audit_AUTHENTICATION(Level level, ServletRequest request, String originIssuer, String subjectIdentity) {
        audit_AUTHENTICATION(level, getRequestTransaction(request), originIssuer, subjectIdentity);
    }

    public void audit_AUTHN_REQ_SENT(Level level, ServletRequest request, String messageId, String destination) {
        audit_AUTHN_REQ_SENT(level, getRequestTransaction(request), messageId, destination);
    }

    public void audit_AUTHN_RESP_SENT(Level level, ServletRequest request, String messageId, String destination, String requestId) {
        audit_AUTHN_RESP_SENT(level, getRequestTransaction(request), messageId, destination, requestId);
    }

    public void audit_AUTHN_REQ_RECEIVED(Level level, ServletRequest request, String messageId, String issuer) {
        audit_AUTHN_REQ_RECEIVED(level, getRequestTransaction(request), messageId, issuer);
    }

    public void audit_AUTHN_RESP_RECEIVED(Level level, ServletRequest request, String messageId, String issuer, String requestId) {
        audit_AUTHN_RESP_RECEIVED(level, getRequestTransaction(request), messageId, issuer, requestId);
    }

    public void audit_ATTR_REQ_SENT(Level level, ServletRequest request, String messageId, String destination) {
        audit_ATTR_REQ_SENT(level, getRequestTransaction(request), messageId, destination);
    }

    public void audit_ATTR_RESP_SENT(Level level, ServletRequest request, String messageId, String destination, String requestId) {
        audit_ATTR_RESP_SENT(level, getRequestTransaction(request), messageId, destination, requestId);
    }

    public void audit_ATTR_REQ_RECEIVED(Level level, ServletRequest request, String messageId, String issuer) {
        audit_ATTR_REQ_RECEIVED(level, getRequestTransaction(request), messageId, issuer);
    }

    public void audit_ATTR_RESP_RECEIVED(Level level, ServletRequest request, String messageId, String issuer, String requestId) {
        audit_ATTR_RESP_RECEIVED(level, getRequestTransaction(request), messageId, issuer, requestId);
    }

    public void audit_AUTHZ_REQ_SENT(Level level, ServletRequest request, String messageId, String destination) {
        audit_AUTHZ_REQ_SENT(level, getRequestTransaction(request), messageId, destination);
    }

    public void audit_AUTHZ_RESP_SENT(Level level, ServletRequest request, String messageId, String destination, String requestId) {
        audit_AUTHZ_RESP_SENT(level, getRequestTransaction(request), messageId, destination, requestId);
    }

    public void audit_AUTHZ_REQ_RECEIVED(Level level, ServletRequest request, String messageId, String issuer) {
        audit_AUTHZ_REQ_RECEIVED(level, getRequestTransaction(request), messageId, issuer);
    }

    public void audit_AUTHZ_RESP_RECEIVED(Level level, ServletRequest request, String messageId, String issuer, String requestId) {
        audit_AUTHZ_RESP_RECEIVED(level, getRequestTransaction(request), messageId, issuer, requestId);
    }

    public void audit_UNKNOWN_MESSAGE_RECEIVED(Level level, ServletRequest request) {
        audit_UNKNOWN_MESSAGE_RECEIVED(level, getRequestTransaction(request));
    }

    public void audit_AUTHZ_PERMITTED_BY_POLICY(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.AUTHZ_PERMITTED_BY_POLICY, "Authorization granted by policy to " + originIssuer + " for attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + ".");
    }

    public void audit_AUTHZ_DENIED_BY_POLICY(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.AUTHZ_DENIED_BY_POLICY, "Authorization denied by policy to " + originIssuer + " for attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + ".");
    }

    public void audit_AUTHZ_PERMITTED_BY_USER(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.AUTHZ_PERMITTED_BY_USER, "Authorization granted by user consent to " + originIssuer + " for attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + ".");
    }

    public void audit_AUTHZ_DENIED_BY_USER(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.AUTHZ_DENIED_BY_USER, "Authorization denied by user consent to " + originIssuer + " for attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + ".");
    }

    public void audit_ATTR_RELEASE(Level level, Transaction transaction, String originIssuer, String attrName, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.ATTR_RELEASE, "Attribute " + attrName + " of user " + subjectIdentity + " released to " + originIssuer + ".");
    }

    public void audit_ATTR_RELEASE(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.ATTR_RELEASE, "Attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + " released to " + originIssuer + ".");
    }

    public void audit_AUTHZ_TOKEN_RELEASE(Level level, Transaction transaction, String originIssuer, List<String> attrNames, String subjectIdentity, Date date) {
        if (attrNames.isEmpty()) {
            audit(level, transaction, SemiramisEventId.AUTHZ_TOKEN_RELEASE, "Authorization token released to " + originIssuer + " of user " + subjectIdentity + " until " + date.toString() + ".");
        } else {
            audit(level, transaction, SemiramisEventId.AUTHZ_TOKEN_RELEASE, "Authorization token released to " + originIssuer + " for attributes " + toAuditMsgString(attrNames) + " of user " + subjectIdentity + " until " + date.toString() + ".");
        }
    }

    public void audit_ATTR_TRANSLATION(Level level, Transaction transaction, String attrName, String newAttrName, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.ATTR_TRANSLATION, "Attribute " + attrName + " of user " + subjectIdentity + " translated to " + newAttrName + ".");
    }

    public void audit_AUTHENTICATION(Level level, Transaction transaction, String originIssuer, String subjectIdentity) {
        audit(level, transaction, SemiramisEventId.AUTHENTICATION, "User " + subjectIdentity + " Authenticated to " + originIssuer + ".");
    }

    public void audit_AUTHN_REQ_SENT(Level level, Transaction transaction, String messageId, String destination) {
        audit(level, transaction, SemiramisEventId.AUTHN_REQ_SENT, "Sent authentication request with id " + messageId + " to " + destination + ".");
    }

    public void audit_AUTHN_RESP_SENT(Level level, Transaction transaction, String messageId, String destination, String requestId) {
        audit(level, transaction, SemiramisEventId.AUTHN_RESP_SENT, "Sent authentication response with id " + messageId + " to " + destination + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_AUTHN_REQ_RECEIVED(Level level, Transaction transaction, String messageId, String issuer) {
        audit(level, transaction, SemiramisEventId.AUTHN_REQ_RECEIVED, "Received authentication request with id " + messageId + " from " + issuer + ".");
    }

    public void audit_AUTHN_RESP_RECEIVED(Level level, Transaction transaction, String messageId, String issuer, String requestId) {
        audit(level, transaction, SemiramisEventId.AUTHN_RESP_RECEIVED, "Received authentication response with id " + messageId + " from " + issuer + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_ATTR_REQ_SENT(Level level, Transaction transaction, String messageId, String destination) {
        audit(level, transaction, SemiramisEventId.ATTR_REQ_SENT, "Sent attribute request with id " + messageId + " to " + destination + ".");
    }

    public void audit_ATTR_RESP_SENT(Level level, Transaction transaction, String messageId, String destination, String requestId) {
        audit(level, transaction, SemiramisEventId.ATTR_RESP_SENT, "Sent attribute response with id " + messageId + " to " + destination + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_ATTR_REQ_RECEIVED(Level level, Transaction transaction, String messageId, String issuer) {
        audit(level, transaction, SemiramisEventId.ATTR_REQ_RECEIVED, "Received attribute request with id " + messageId + " from " + issuer + ".");
    }

    public void audit_ATTR_RESP_RECEIVED(Level level, Transaction transaction, String messageId, String issuer, String requestId) {
        audit(level, transaction, SemiramisEventId.ATTR_RESP_RECEIVED, "Received attribute response with id " + messageId + " from " + issuer + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_AUTHZ_REQ_SENT(Level level, Transaction transaction, String messageId, String destination) {
        audit(level, transaction, SemiramisEventId.AUTHZ_REQ_SENT, "Sent Authorization request with id " + messageId + " to " + destination + ".");
    }

    public void audit_AUTHZ_RESP_SENT(Level level, Transaction transaction, String messageId, String destination, String requestId) {
        audit(level, transaction, SemiramisEventId.AUTHZ_RESP_SENT, "Sent Authorization response with id " + messageId + " to " + destination + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_AUTHZ_REQ_RECEIVED(Level level, Transaction transaction, String messageId, String issuer) {
        audit(level, transaction, SemiramisEventId.AUTHZ_REQ_RECEIVED, "Received Authorization request with id " + messageId + " from " + issuer + ".");
    }

    public void audit_AUTHZ_RESP_RECEIVED(Level level, Transaction transaction, String messageId, String issuer, String requestId) {
        audit(level, transaction, SemiramisEventId.AUTHZ_RESP_RECEIVED, "Received Authorization response with id " + messageId + " from " + issuer + (iaConfig.DISABLE_MSG_CORRELATION_AUDIT ? "" : " in response to request with id " + requestId) + ".");
    }

    public void audit_UNKNOWN_MESSAGE_RECEIVED(Level level, Transaction transaction) {
        audit(level, transaction, SemiramisEventId.AUTHZ_RESP_RECEIVED, "Unknown message received.");
    }

    private String toAuditMsgString(List<String> attrNames) {
        String auditMsgString;
        Iterator<String> attrNamesIterator = attrNames.iterator();
        if (attrNamesIterator.hasNext()) {
            auditMsgString = attrNamesIterator.next();
        } else {
            throw new IllegalArgumentException("No attributes specified.");
        }
        while (attrNamesIterator.hasNext()) {
            auditMsgString = auditMsgString + ", " + attrNamesIterator.next();
        }
        return auditMsgString;
    }
}
