/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.Audit.Level;
import eu.semiramis.identityaggregator.PDP.Action;
import eu.semiramis.identityaggregator.commons.Attribute;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import javax.servlet.http.HttpServletRequest;
import org.opensaml.xacml.ctx.DecisionType.DECISION;

/**
 * Identity Aggregator - AuthzProvider Module
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class AuthzProvider {

    protected IdentityAggregator_Config config;
    protected Audit audit;
    protected PDP pdp;
    protected static final String AUTHENTICATION_REQUEST = "AuthenticationRequest";
    protected static final String AUTHORIZATION_REQUEST = "AuthorizationRequest";
    protected static final String ATTRIBUTE_REQUEST = "AttributeRequest";

    public AuthzProvider(IdentityAggregator_Config config, Transaction transaction) {
        this.config = config;
        audit = new Audit(AuthzProvider.class.getName());
        audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.AUTHORIZATION_INFO, "Instantiating PDP module...");
        pdp = PDPFactory.newPDPInstance(config);
        audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.AUTHORIZATION_INFO, "Done.");
    }

    public Decision isAuthenticationAllowed(HttpServletRequest request, String issuer, String originIssuer, Destination destination) {
        String target = destination.getDomain();
        if (target == null) {
            target = destination.getProviderId();
        }
        return evaluate(request, Action.AUTHENTICATION, issuer, originIssuer, target, null, null);
    }

    public Map<String, Decision> isAttributeQueryAllowed(HttpServletRequest request, String issuer, String requestor, String subject, Map<Destination, List<Attribute>> providersMap) {
        Map<String, Decision> result = new HashMap<String, Decision>();
        boolean allowed = true;
        List<String> attributes = new ArrayList<String>();

        for (Map.Entry<Destination, List<Attribute>> entry : providersMap.entrySet()) {
            for (Attribute attribute : entry.getValue()) {
                String attributeId = attribute.getId();
                attributes.add(attributeId);
                for (String attrGroup : getGroups(request, attributeId)) {
                    String target = entry.getKey().getDomain();

                    if (target == null) {
                        target = entry.getKey().getProviderId();
                    }
                    Decision decision = evaluate(request, Action.ATTRIBUTE_RELEASE, issuer, requestor, target, subject, attrGroup);
                    result.put(attributeId, decision);
                    allowed &= decision.getDecision().equals(DECISION.Permit);
                }
            }
        }
        if (allowed) {
            audit.audit_AUTHZ_PERMITTED_BY_POLICY(Level.INFO, request, requestor, attributes, subject);
        } else {
            audit.audit_AUTHZ_DENIED_BY_POLICY(Level.INFO, request, requestor, attributes, subject);
        }
        return result;
    }

    public Decision isAuthorizationAllowed(HttpServletRequest request, String issuer, String originIssuer, Destination destination) {
        String target = destination.getDomain();
        if (target == null) {
            target = destination.getProviderId();
        }
        return evaluate(request, Action.AUTHORIZATION, issuer, originIssuer, target, null, null);
    }

    private Decision evaluate(HttpServletRequest request, Action action, String issuer, String requestor, String target, String subject, String resource) {
        try {
            audit.audit(Level.DEBUG, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Asking PDP for authorization...");
            Decision decision = pdp.evaluate(request, action, issuer, requestor, target, subject, resource);
            if (decision.getDecision() == DECISION.Permit) {
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Is allowed");
            } else {
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Is not allowed");
            }
            return decision;
        } catch (Exception ex) {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.AUTHORIZATION_INFO, ex);
            return new Decision(DECISION.Indeterminate);
        }
    }

    private List<String> getAttributesGroups(HttpServletRequest request, List<Attribute> attributes) {
        List<String> groups = new ArrayList<String>();
        Iterator<Attribute> it = attributes.iterator();
        while (it.hasNext()) {
            String attribute = it.next().getId();
            List<String> group = getGroups(request, attribute);
            Iterator<String> it2 = group.iterator();
            while (it2.hasNext()) {
                String groupName = it2.next();
                if (!groups.contains(groupName)) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Setting group for [" + attribute + "] : " + groupName);
                    groups.add(groupName);
                }
            }

        }
        return groups;
    }

    private List<String> getGroups(HttpServletRequest request, String attribute) {
        if (config.AttributeGroupMap.containsKey(attribute)) {
            audit.audit(Level.INFO, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Attribute group found for attribute [" + attribute + "], using group for policy evaluation");
            return config.AttributeGroupMap.get(attribute);
        } else {
            List<String> nonGroup = new ArrayList();
            nonGroup.add(attribute);
            return nonGroup;
        }
    }
}
