/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.Audit.Level;
import eu.semiramis.identityaggregator.commons.AttributeToken;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.identityaggregator.commons.Provider;
import eu.semiramis.identityaggregator.commons.Token;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import org.joda.time.DateTime;
import org.opensaml.DefaultBootstrap;
import org.opensaml.common.impl.RandomIdentifierGenerator;
import org.opensaml.saml2.binding.decoding.HTTPPostDecoder;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;

/**
 * Identity Aggregator Server Component - Response processor
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IAConsumerService {

    protected IdentityAggregator_Config config;
    protected ID_Discovery idDiscovery;
    protected AuthzProvider authzProvider;
    protected Audit audit;
    protected TrustManager trustManager;
    protected AttributeTranslator attributeTranslator;
    protected ID_Translator idTranslator;

    public IAConsumerService(String configFilePath, Transaction transaction) throws Exception {
        config = new IdentityAggregator_Config(configFilePath);
        init(transaction);
    }

    public ResponseMessage processResponse(HttpServletRequest request)
            throws ServletException, IOException {
        Response samlResponse = null;
        try {
            samlResponse = (Response) SAMLUtil.decodeSAML(request, new HTTPPostDecoder());
        } catch (Exception ex) {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
            return null;
        }
        audit.audit(Level.INFO, request, Audit.AppLogEventId.RESPONSE_PROCESSING_INFO, "SAMLResponse received:" + SAMLUtil.SAMLtoString(samlResponse));
        SAMLResponseMessage samlResponseMessage = new SAMLResponseMessage(samlResponse);
        if (!trustManager.validateSign(request, samlResponse, samlResponse.getIssuer().getValue())) {
            samlResponse.getStatus().getStatusCode().setValue(StatusCode.REQUEST_DENIED_URI);
            samlResponse.getStatus().getStatusMessage().setMessage("Signature not valid.");
        }

        samlResponseMessage.setCredentialsInfo(config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
        return samlResponseMessage;
    }

    public ID_Discovery getDiscovery() {
        return idDiscovery;
    }

    public AuthzProvider getAuthzProvider() {
        return authzProvider;
    }

    public AttributeTranslator getAttributeTranslator() {
        return attributeTranslator;
    }

    public TrustManager getTrustManager() {
        return trustManager;
    }

    public Audit getAudit() {
        return audit;
    }

    public IdentityAggregator_Config getConfig() {
        return config;
    }

    /*
    protected List<Assertion> getAssertions(List<Token> tokens){
    List<Assertion> assertions = new ArrayList<Assertion>();
    Iterator<Token> it = tokens.iterator();
    while (it.hasNext()) {
    Token token = it.next();
    Assertion assertion = token.getAssertion();
    assertions.add(assertion);
    }
    return assertions;
    }
    
    protected void processAttrResponse(AttributeQuery incomingAttrQuery, List<Token> tokens, HttpServletResponse response) throws Exception {
    String consumerURL = RequestLinker.getAssociateConsumerService(incomingAttrQuery.getID());
    Response outcomingResponse = SAMLUtil.createResponse(config.SEMIRAMIS_IA_ID, incomingAttrQuery.getID(), consumerURL, getAssertions(tokens));
    if(config.SIGN_MESSAGES){
    SAMLUtil.signXML(outcomingResponse, config.API_PKCS12_PATH, config.API_PKCS12_PASSWORD, config.API_PKCS12_CERT_ALIAS, config.API_PRIVATE_KEY_PASSWORD);
    }
    SAMLUtil.httpPost(response, consumerURL, outcomingResponse, consumerURL);
    }
     */
    protected void init(Transaction transaction) {
        try {
            Audit.initAudit(config);
            audit = new Audit(IAConsumerService.class.getName());
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Instantiating ID_Discory module...");
            idDiscovery = new ID_Discovery(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Instantiating AuthzProvider module...");
            authzProvider = new AuthzProvider(config, transaction);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Instantiating TrustManager module...");
            trustManager = new TrustManager(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Instantianting AttributeTranslator module...");
            attributeTranslator = new AttributeTranslator(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Instantiating ID_Translator module...");
            idTranslator = new ID_Translator(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_INFO, "Done.");
            // Start openSAML library
            DefaultBootstrap.bootstrap();
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    List<Token> generateTokens(HttpServletRequest request, List<Token> tokens, String destination) {
        Provider provider = config.Providers.get(destination);
        if (provider.isGenerateNewTokens()) {
            List<Token> newTokens = new ArrayList<Token>();
            Iterator<Token> it = tokens.iterator();
            while (it.hasNext()) {
                Token token = it.next();
                Token newToken = null;
                try {
                    if (token instanceof AuthenticationToken) {
                        newToken = new AuthenticationToken(token.getAssertion());
                    } else if (token instanceof AttributeToken) {
                        newToken = new AttributeToken(token.getAssertion());
                    } else if (token instanceof AuthorizationToken) {
                        newToken = new AuthorizationToken(token.getAssertion());
                    } else {
                        audit.audit(Level.WARN, request, Audit.AppLogEventId.TOKEN_PROCESSING_ERROR, "Token not recognized");
                        continue;
                    }
                    newToken.getAssertion().getIssuer().setValue(config.ID);
                    newToken.getAssertion().setID(new RandomIdentifierGenerator().generateIdentifier());
                    newToken.getAssertion().setIssueInstant(new DateTime());
                    if (provider.isUsePseudonyms()) {
                        String pseudonym = ID_Translator.generatePseudonym();
                        ID_Translator.linkPseudonym(pseudonym, token.getAssertion().getSubject().getNameID().getValue());
                        newToken.getAssertion().getSubject().getNameID().setValue(pseudonym);
                    }
                    newTokens.add(newToken);
                } catch (Exception ex) {
                    audit.audit(Level.ERROR, request, Audit.AppLogEventId.TOKEN_PROCESSING_ERROR, ex);
                }
            }
            return newTokens;
        } else {
            return tokens;
        }
    }

    ID_Translator getID_Translator() {
        return idTranslator;
    }
}
