/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.Audit.Level;
import java.io.IOException;

import javax.servlet.*;
import javax.servlet.http.*;

import org.opensaml.DefaultBootstrap;
import org.opensaml.common.binding.decoding.SAMLMessageDecoder;
import org.opensaml.saml2.binding.decoding.HTTPPostDecoder;
import org.opensaml.saml2.binding.decoding.HTTPRedirectDeflateDecoder;
import org.opensaml.saml2.core.AuthnRequest;


import org.opensaml.saml2.core.RequestAbstractType;
import org.opensaml.saml2.core.StatusCode;

/**
 * Identity Aggregator Server Component
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IARequestProcessor {

    protected IdentityAggregator_Config config;
    protected ID_Discovery idDiscovery;
    protected AuthzProvider authzProvider;
    protected Audit audit;
    protected AttributeTranslator attributeTranslator;
    protected TrustManager trustManager;
    protected ID_Translator idTranslator;

    public ID_Discovery getDiscovery() {
        return idDiscovery;
    }

    public AuthzProvider getAuthzProvider() {
        return authzProvider;
    }

    public IdentityAggregator_Config getConfig() {
        return config;
    }

    public Audit getAudit() {
        return audit;
    }

    public AttributeTranslator getAttributeTranslator() {
        return attributeTranslator;
    }

    public IARequestProcessor(String configFilePath, Transaction transaction) throws Exception {
        config = new IdentityAggregator_Config(configFilePath);
        init(transaction);
    }

    public RequestMessage processRequest(HttpServletRequest request) {
        SAMLMessageDecoder decoder = null;
        if (request.getMethod().equals("GET")) {
            decoder = new HTTPRedirectDeflateDecoder();
        } else if (request.getMethod().equals("POST")) {
            decoder = new HTTPPostDecoder();
        } else {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, "Request method not supported: " + request.getMethod());
            return null;
        }
        if (request.getParameter("SAMLRequest") != null) {
            try {
                // There exists a SAMLRequest which needs to be processed
                RequestAbstractType samlRequest = (RequestAbstractType) SAMLUtil.decodeSAML(request, decoder);
                audit.audit(Level.INFO, request, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "SAMLRequest received:" + SAMLUtil.SAMLtoString(samlRequest));
                SAMLRequestMessage requestMessage = new SAMLRequestMessage(samlRequest);
                String consumerService = request.getParameter("RelayState");
                if (consumerService == null && samlRequest instanceof AuthnRequest) {
                    consumerService = ((AuthnRequest) samlRequest).getAssertionConsumerServiceURL();
                }
                requestMessage.setConsumerServiceURL(consumerService);
                // check signature
                String statusCode = StatusCode.SUCCESS_URI;
                String statusMessage = null;
                if (!trustManager.validateSign(request, samlRequest, samlRequest.getIssuer().getValue())) {
                    statusCode = StatusCode.REQUEST_DENIED_URI;
                    statusMessage = "Signature not valid.";
                    audit.audit(Level.WARN, request, Audit.AppLogEventId.INVALID_SIGNATURE, "Cannot validate signature");
                }
                requestMessage.setStatusCode(statusCode);
                if (statusMessage != null) {
                    requestMessage.setStatusMessage(statusMessage);
                }
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Consumer Service of received request = [" + consumerService + "]");
                return requestMessage;
            } catch (Exception ex) {
                audit.audit(Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
            }
        }
        return null;
    }

    protected void init(Transaction transaction) {
        try {
            Audit.initAudit(config);
            audit = new Audit(IARequestProcessor.class.getName());
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Instantiating ID_Discory module...");
            idDiscovery = new ID_Discovery(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Instantiating AuthzProvider module...");
            authzProvider = new AuthzProvider(config, transaction);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Instantiating AttributeTranslator module...");
            attributeTranslator = new AttributeTranslator(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Instantiating TrustManager module...");
            trustManager = new TrustManager(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Instantiating ID_Translator module...");
            idTranslator = new ID_Translator(config);
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
            // Start openSAML library
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Initializing OpenSAML libraries...");
            DefaultBootstrap.bootstrap();
            audit.audit(Level.DEBUG, transaction, Audit.AppLogEventId.IA_INITIALIZATION_INFO, "Done.");
        } catch (Exception ex) {
            ex.printStackTrace();
        }
    }

    ID_Translator getID_Translator() {
        return idTranslator;
    }

    public void showErrorWeb(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        RequestDispatcher rd = request.getRequestDispatcher("error.jsp");
        rd.include(request, response);
    }
}
