/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import java.util.AbstractMap.SimpleEntry;
import java.util.HashMap;
import java.util.Map;
import java.util.Map.Entry;
import javax.servlet.http.HttpServletRequest;
import org.opensaml.common.impl.RandomIdentifierGenerator;

/**
 * Identity Aggregator - Id Translator Module
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class ID_Translator {

    private static Map<String, String> pseudonymLinker = new HashMap<String, String>();
    private IdentityAggregator_Config config;
    Audit audit;

    public static String generatePseudonym() {
        return new RandomIdentifierGenerator().generateIdentifier();
    }

    public static void linkPseudonym(String pseudonym, String id) {
        pseudonymLinker.put(pseudonym, id);
    }

    public static String getRealId(String pseudonym) {
        return pseudonymLinker.get(pseudonym);
    }

    public ID_Translator(IdentityAggregator_Config config) {
        this.config = config;
        audit = new Audit(ID_Translator.class.getName());
    }

    public String translateRequestSubjectId(HttpServletRequest request, String issuer, String destination, String subjectId) {
        return translateRequestSubjectId(issuer, destination, subjectId, audit.getRequestTransaction(request));
    }

    public String translateRequestSubjectId(String issuer, String destination, String subjectId, Transaction transaction) {
        audit.audit(Audit.Level.INFO, transaction, Audit.AppLogEventId.ID_TRANSLATION_INFO, "Translating identities from " + issuer + " to " + destination);
        Entry<String, String> entry = new SimpleEntry<String, String>(issuer, destination);
        HashMap<String, String> identityDictionary = config.IdentityDictionariesMap.get(entry);
        if (identityDictionary == null) {
            audit.audit(Audit.Level.WARN, transaction, Audit.AppLogEventId.ID_TRANSLATION_ERROR, "No dictionary configured");
            return subjectId;
        }

        String translatedSubjectId;
        if (identityDictionary.containsKey(subjectId)) {
            translatedSubjectId = identityDictionary.get(subjectId);
        } else {
            translatedSubjectId = subjectId;
        }
        audit.audit(Audit.Level.DEBUG, transaction, Audit.AppLogEventId.ID_TRANSLATION_INFO, "[identity]" + subjectId + "=" + translatedSubjectId);

        return translatedSubjectId;
    }

    public String translateResponseSubjectId(HttpServletRequest request, String destination, String issuer, String subjectId) {
        return translateResponseSubjectId(destination, issuer, subjectId, audit.getRequestTransaction(request));
    }

    public String translateResponseSubjectId(String destination, String issuer, String subjectId, Transaction transaction) {
        audit.audit(Audit.Level.INFO, transaction, Audit.AppLogEventId.ID_TRANSLATION_INFO, "Translating identities from " + issuer + " to " + destination);
        Entry<String, String> entry = new SimpleEntry<String, String>(destination, issuer);
        HashMap<String, String> identityDictionary = config.IdentityDictionariesMap.get(entry);
        if (identityDictionary == null) {
            audit.audit(Audit.Level.WARN, transaction, Audit.AppLogEventId.ID_TRANSLATION_ERROR, "No dictionary configured");
            return subjectId;
        }

        String translatedSubjectId;
        if (identityDictionary.containsKey(subjectId)) {
            translatedSubjectId = identityDictionary.get(subjectId);
        } else {
            translatedSubjectId = subjectId;
        }
        audit.audit(Audit.Level.DEBUG, transaction, Audit.AppLogEventId.ID_TRANSLATION_INFO, "[identity]" + subjectId + "=" + translatedSubjectId);

        return translatedSubjectId;
    }
}
