/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.identityaggregator.Audit.Level;
import java.io.FileInputStream;
import java.io.IOException;
import javax.servlet.http.HttpServletRequest;
import org.opensaml.xacml.ctx.DecisionType.DECISION;
import org.opensaml.xml.parse.BasicParserPool;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Identity Aggregator - Internal Policy Decision Point (PDP)
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class InternalPDP implements PDP {

    protected IdentityAggregator_Config config;
    protected TrustManager trustManager;
    private Audit audit;
    static private final String USE_IDEAS = "USE_IDEAS";

    InternalPDP(IdentityAggregator_Config config) {
        this.config = config;
        audit = new Audit(InternalPDP.class.getName());
        trustManager = new TrustManager(config);

    }

    @Override
    public Decision evaluate(HttpServletRequest request, Action action, String issuer, String requestor, String target, String subject, String resource) throws PDPException {
        return evaluate(request, subject, action, resource, null);
    }

    public Decision evaluate(HttpServletRequest request, String subject, Action action, String resource, String role) throws PDPException {
        audit.audit(Level.INFO, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "evaluate authorization: [Subject=" + subject + ", Action=" + action + ", Resource=" + resource + ", Role=" + role + "]");
        Decision decision = evaluateInternalPolicy(request, config.POLICY_MGR_CONFIG_FILE_PATH, subject, action, resource, role);
        if (decision.getObligations() != null && decision.getObligations().contains(Decision.Obligation.USE_IDEAS)) {
            try {
                audit.audit(Level.INFO, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Obligation suggest using IDEAS, trying with the external Authorization Providers");
                decision = trustManager.evaluate(request, subject, action.name(), resource, role);
                audit.audit(Level.INFO, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Decision: " + decision);
                return decision;
            } catch (Exception ex) {
                throw new PDPException(ex.getMessage(), ex);
            }
        }
        audit.audit(Level.INFO, request, Audit.AppLogEventId.AUTHORIZATION_INFO, "Decision: Permit");
        return new Decision(DECISION.Permit);
    }

    /**
     * Evaluates policies contanined in an internal file
     * @param policyFilePath
     * @param subject
     * @param action
     * @param resource
     * @param role
     * @return
     */
    private Decision evaluateInternalPolicy(HttpServletRequest request, String policyFilePath, String subject, Action action, String resource, String role) {

        Decision decision = new Decision(DECISION.Indeterminate);
        audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Evaluating internal policies");
        if (policyFilePath == null) {
            audit.audit(Level.WARN, request, Audit.AppLogEventId.INTERNAL_POLICIES_NOT_FOUND, "Internal policies not found");
            return decision;
        }
        FileInputStream f = null;
        try {
            BasicParserPool b = new BasicParserPool();
            f = new FileInputStream(policyFilePath);
            Document d = b.parse(f);

            NodeList nodeList = d.getElementsByTagName("Rule");
            for (int i = 0; i < nodeList.getLength(); i++) {
                String ruleSubject = null;
                String ruleAction = null;
                String ruleEffect = null;
                String ruleObligation = null;

                Node node = nodeList.item(i).getAttributes().getNamedItem("Subject");
                if (node != null) {
                    ruleSubject = node.getNodeValue();
                }
                node = nodeList.item(i).getAttributes().getNamedItem("Action");
                if (node != null) {
                    ruleAction = node.getNodeValue();
                }
                node = nodeList.item(i).getAttributes().getNamedItem("Effect");
                if (node != null) {
                    ruleEffect = node.getNodeValue();
                }
                node = nodeList.item(i).getAttributes().getNamedItem("Obligation");
                if (node != null) {
                    ruleObligation = node.getNodeValue();
                }

                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "evaluating rule: Subject=[" + ruleSubject + "], Action=[" + ruleAction + "], Effect[" + ruleEffect + "], Obligation=[" + ruleObligation + "]");

                if (ruleSubject != null && ((subject == null) || !subject.equals(ruleSubject))) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "rule NOT MATCH");
                    continue;
                }
                String actionStr = action.name();
                if (action.equals(Action.ATTRIBUTE_RELEASE)) {
                    actionStr = actionStr + "/" + resource;
                }
                if (ruleAction != null && !actionStr.equals(ruleAction)) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "rule NOT MATCH");
                    continue;
                }

                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "rule MATCH");

                decision.setDecision(DECISION.valueOf(ruleEffect));
                if (ruleObligation != null) {
                    if (ruleObligation.equals(USE_IDEAS)) {
                        decision.getObligations().add(Decision.Obligation.USE_IDEAS);
                    }
                }

                return decision;
            }

        } catch (Exception ex) {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.POLICY_EVALUATION_ERROR, ex);
        } finally {
            try {
                f.close();
            } catch (IOException ex) {
                audit.audit(Level.ERROR, request, Audit.AppLogEventId.POLICY_EVALUATION_ERROR, ex);
            }
        }
        return decision;
    }
}
