/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.identityaggregator.Audit.Level;
import eu.semiramis.policyManager.pdp.ContextHandler;
import eu.semiramis.policyManager.pdp.impl.ContextHandlerFactory;
import eu.semiramis.policymodel.request.AttributeReleaseRequest;
import eu.semiramis.policymodel.request.ServiceToServiceRequest;
import eu.semiramis.policymodel.response.AttributeReleaseResponseHandler;
import eu.semiramis.policymodel.response.ServiceToServiceResponseHandler;
import java.io.FileInputStream;
import java.util.Properties;
import javax.servlet.http.HttpServletRequest;
import org.opensaml.xacml.ctx.DecisionType.DECISION;
import org.opensaml.xacml.ctx.ResponseType;

/**
 * PolicyMgmtPDP
 * 
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class PolicyMgmtPDP implements PDP {

    private IdentityAggregator_Config config;
    private Audit audit;

    public PolicyMgmtPDP(IdentityAggregator_Config config) {
        this.config = config;
        audit = new Audit(PolicyMgmtPDP.class.getName());
    }

    /**
     * Evaluates the policies to determine whether the request is authorized or not
     * @param request the servlet request.
     * @param action the requested action
     * @param issuer the immediate issuer of the request
     * @param requestor the origin issuer of the request
     * @param target the target service that will serve the request
     * @param subject the id of the subject whose resource is requested
     * @param resource the resource that is being requested
     * @return a decision
     * @throws PDPException 
     */
    @Override
    public Decision evaluate(HttpServletRequest request, Action action, String issuer, String requestor, String target, String subject, String resource) throws PDPException {
        try {
            Properties properties = new Properties();
            properties.load(new FileInputStream(config.POLICY_MGR_CONFIG_FILE_PATH));
            ContextHandler handler = ContextHandlerFactory.createContextHandler(properties);

            if (config.ENABLE_TRUST_POLICIES) {
                // Check whether the request comes from a trusted service
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Checking whether the request comes from a trusted service...");
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Issuing ServiceToService XACML request for " + action.name() + " from " + issuer + " to " + config.ID);
                ServiceToServiceRequest pdpRequest = new ServiceToServiceRequest(issuer, action.name(), config.ID);
                org.opensaml.xacml.ctx.RequestType xacmlRequest = pdpRequest.toXACMLRequest();
                ResponseType xacmlResponse = handler.handleRequest(xacmlRequest);
                ServiceToServiceResponseHandler response = new ServiceToServiceResponseHandler();
                response.handle(xacmlResponse);
                if (!response.isAuthorized()) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request denied.");
                    return new Decision(DECISION.Deny);
                }
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request permitted.");

                // Check whether outgoing request goes to a trusted service
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Checking whether outgoing request goes to a trusted service...");
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Issuing ServiceToService XACML request for " + action.name() + " from " + config.ID + " to " + target);
                pdpRequest = new ServiceToServiceRequest(config.ID, action.name(), target);
                xacmlRequest = pdpRequest.toXACMLRequest();
                xacmlResponse = handler.handleRequest(xacmlRequest);
                response = new ServiceToServiceResponseHandler();
                response.handle(xacmlResponse);
                if (!response.isAuthorized()) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request denied.");
                    return new Decision(DECISION.Deny);
                }
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request permitted.");
            }

            // Check whether the request is allowed from the origin issuer to the target service
            audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Checking whether the request is allowed from the origin issuer to the target service...");
            if (action.equals(Action.AUTHENTICATION) || action.equals(Action.AUTHORIZATION)) {
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Issuing ServiceToService XACML request for " + action.name() + " from " + requestor + " to " + target);
                ServiceToServiceRequest pdpRequest = new ServiceToServiceRequest(requestor, action.name(), target);
                org.opensaml.xacml.ctx.RequestType xacmlRequest = pdpRequest.toXACMLRequest();
                ResponseType xacmlResponse = handler.handleRequest(xacmlRequest);
                ServiceToServiceResponseHandler response = new ServiceToServiceResponseHandler();
                response.handle(xacmlResponse);
                Decision decision;
                if (response.isAuthorized()) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request permitted.");
                    decision = new Decision(DECISION.Permit);
                } else {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request denied.");
                    decision = new Decision(DECISION.Deny);
                }
                return decision;
            } else if (action.equals(Action.ATTRIBUTE_RELEASE)) {
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Issuing AttributeRelease XACML request for attribute " + resource + " of subject " + subject + " from " + requestor + " to " + target);
                AttributeReleaseRequest pdpRequest = new AttributeReleaseRequest(requestor, target, subject, resource);
                org.opensaml.xacml.ctx.RequestType xacmlRequest = pdpRequest.toXACMLRequest();
                ResponseType xacmlResponse = handler.handleRequest(xacmlRequest);
                AttributeReleaseResponseHandler response = new AttributeReleaseResponseHandler();
                response.handle(xacmlResponse);
                Decision decision;
                if (response.isAuthorized()) {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request permitted.");
                    decision = new Decision(DECISION.Permit);
                } else {
                    audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "Request denied.");
                    decision = new Decision(DECISION.Deny);
                }
                if (response.getConsentRequired()) {
                    decision.getObligations().add(Decision.Obligation.ASK_FOR_USER_CONSENT);
                }
                if (response.getNotificationRequired()) {
                    decision.getObligations().add(Decision.Obligation.NOTIFY_USER);
                }
                return decision;
            } else {
                audit.audit(Level.ERROR, request, Audit.AppLogEventId.POLICY_EVALUATION_ERROR, "Unsupported action type: " + action);
                return new Decision(DECISION.NotApplicable);
            }
        } catch (Exception e) {
            throw new PDPException(e.getMessage(), e);
        }
    }
}
