/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLAttribute;
import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.identityaggregator.commons.Attribute;
import eu.semiramis.identityaggregator.commons.AttributeToken;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.identityaggregator.commons.Conditions;
import eu.semiramis.identityaggregator.commons.Token;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import org.opensaml.saml2.common.Extensions;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.AttributeQuery;
import org.opensaml.saml2.core.Audience;
import org.opensaml.saml2.core.AudienceRestriction;
import org.opensaml.saml2.core.AuthnRequest;
import org.opensaml.saml2.core.AuthzDecisionQuery;
import org.opensaml.saml2.core.IDPEntry;
import org.opensaml.saml2.core.Issuer;
import org.opensaml.saml2.core.RequestAbstractType;
import org.opensaml.saml2.core.Scoping;
import org.opensaml.saml2.core.Subject;
import org.opensaml.saml2.core.SubjectQuery;
import org.opensaml.xml.XMLObject;

/**
 * SAMLRequestMessage
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class SAMLRequestMessage extends RequestMessage {

    private RequestAbstractType samlRequest;
    private List<Token> tokens = null;
    private String consumerServiceURL;
    private String destinationId;
    private String statusCode;
    private String statusMessage;

    SAMLRequestMessage(RequestAbstractType samlRequest) {
        this.samlRequest = samlRequest;
    }

    public RequestAbstractType getSamlRequest() {
        return samlRequest;
    }

    @Override
    public String getId() {
        return samlRequest.getID();
    }

    @Override
    public List<Token> getTokens() {
        if (tokens == null) {
            tokens = new ArrayList<Token>();
            if (samlRequest.getExtensions() == null) {
                return tokens;
            }
            List<Assertion> assertions = SAMLUtil.recoverAssertions(samlRequest.getExtensions());
            Iterator<Assertion> it = assertions.iterator();
            while (it.hasNext()) {
                Assertion assertion = it.next();
                if (!assertion.getAuthnStatements().isEmpty()) {
                    try {
                        AuthenticationToken authenticationToken = new AuthenticationToken(assertion);
                        tokens.add(authenticationToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
                if (!assertion.getAttributeStatements().isEmpty()) {
                    try {
                        AttributeToken attributeToken = new AttributeToken(assertion);
                        tokens.add(attributeToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
                if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                    try {
                        AuthorizationToken authorizationToken = new AuthorizationToken(assertion);
                        tokens.add(authorizationToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
        return tokens;
    }

    @Override
    public List<Attribute> getRequestedAttributes() {
        if (samlRequest instanceof AttributeQuery) {
            List<Attribute> attributes = new ArrayList<Attribute>();
            for (SAMLAttribute samlAttribute : SAMLUtil.getQueryAttributes((AttributeQuery) samlRequest)) {
                attributes.add(new Attribute(samlAttribute));
            }
            return attributes;
        }
        return null;
    }

    @Override
    public RequestType getType() {
        String type = samlRequest.getElementQName().getLocalPart();
        if (type.equals("AuthnRequest")) {
            return RequestType.AUTHENTICATION;
        } else if (type.equals("AttributeQuery")) {
            return RequestType.ATTRIBUTE;
        } else if (type.equals("AuthzDecisionQuery")) {
            return RequestType.AUTHORIZATION;
        }
        return null;

    }

    @Override
    public String getConsumerServiceURL() {
        return consumerServiceURL;
    }

    public void setConsumerServiceURL(String consumerServiceURL) {
        this.consumerServiceURL = consumerServiceURL;
    }

    @Override
    public String getIssuer() {
        return samlRequest.getIssuer().getValue();
    }

    @Override
    public String getOriginIssuer() {
        Extensions extensions = samlRequest.getExtensions();
        if (extensions != null) {
            List<XMLObject> xmlObjects = extensions.getUnknownXMLObjects();
            if (xmlObjects != null) {
                for (XMLObject xmlObject : xmlObjects) {
                    if (xmlObject instanceof Issuer) {
                        return ((Issuer) xmlObject).getValue();
                    }
                }
            }
        }
        return null;
    }

    @Override
    public X509Certificate getAttrEncrX509Cert() throws CertificateException {
        Extensions extensions = samlRequest.getExtensions();
        if (extensions != null) {
            return SAMLUtil.recoverX509Certificate(samlRequest);
        }
        return null;
    }

    @Override
    public String getSubjectId() {
        if (samlRequest instanceof SubjectQuery) {
            return ((SubjectQuery) samlRequest).getSubject().getNameID().getValue();
        }
        return null;
    }

    @Override
    public void changeSubjectId(String subjectId) {
        if (samlRequest instanceof SubjectQuery) {
            Subject subject = ((SubjectQuery) samlRequest).getSubject();
            if (subject != null) {
                subject.getNameID().setValue(subjectId);
                return;
            }
        }
        throw new IllegalStateException("The request does not contain a subjectId to be changed.");
    }

    @Override
    public String getActionId() {
        if (samlRequest instanceof AuthzDecisionQuery) {
            return ((AuthzDecisionQuery) samlRequest).getActions().get(0).getAction();
        }
        return null;
    }

    @Override
    public String getResourceId() {
        if (samlRequest instanceof AuthzDecisionQuery) {
            return ((AuthzDecisionQuery) samlRequest).getResource();
        }
        return null;
    }

    @Override
    public Conditions getConditions() {
        if (samlRequest instanceof AuthnRequest) {
            Conditions conditions = new Conditions();
            Scoping scoping = ((AuthnRequest) samlRequest).getScoping();
            if (scoping != null) {
                if (scoping.getIDPList() != null && scoping.getIDPList().getIDPEntrys() != null) {
                    List<IDPEntry> idpEntries = scoping.getIDPList().getIDPEntrys();
                    if (!idpEntries.isEmpty()) {
                        conditions.setTargetDomain(idpEntries.get(0).getProviderID());
                    }
                }
            }

            org.opensaml.saml2.core.Conditions samlConditions = ((AuthnRequest) samlRequest).getConditions();
            if (samlConditions != null) {
                conditions.setNotBefore(samlConditions.getNotBefore());
                conditions.setNotOnOrAfter(samlConditions.getNotOnOrAfter());
                if (samlConditions.getOneTimeUse() != null) {
                    conditions.setOneTimeUse(true);
                }
                if (samlConditions.getAudienceRestrictions() != null) {
                    List<String> audienceRestrictionList = new ArrayList<String>();
                    Iterator<AudienceRestriction> it = samlConditions.getAudienceRestrictions().iterator();
                    while (it.hasNext()) {
                        AudienceRestriction samlAudienceRestriction = it.next();
                        if (samlAudienceRestriction.getAudiences() != null) {
                            Iterator<Audience> it2 = samlAudienceRestriction.getAudiences().iterator();
                            while (it2.hasNext()) {
                                Audience audience = it2.next();
                                audienceRestrictionList.add(audience.getAudienceURI());
                            }
                        }
                    }
                    conditions.setAudiencieRestriction(audienceRestrictionList);
                }
            }
            return conditions;
        }
        return null;
    }

    public void setStatusCode(String statusCode) {
        this.statusCode = statusCode;
    }

    @Override
    public String getStatusCode() {
        return statusCode;
    }

    public void setStatusMessage(String statusMessage) {
        this.statusMessage = statusMessage;
    }

    @Override
    public String getStatusMessage() {
        return statusMessage;
    }

    /**
     * Methods kept for backward compatibility with (id, value) only attributes.
     */
    /**
     * @deprecated Replaced by getRequestedAttributes().
     */
    @Override
    public List<String> getAttributes() {
        return generateAttributeNames(getRequestedAttributes());
    }

    private List<String> generateAttributeNames(List<Attribute> attributes) {
        if (attributes == null) {
            return null;
        }
        List<String> attributeNames = new ArrayList<String>();
        for (Attribute attribute : attributes) {
            attributeNames.add(attribute.getId());
        }
        return attributeNames;
    }
}
