/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.identityaggregator.commons.Attribute;
import eu.semiramis.identityaggregator.commons.AttributeToken;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.identityaggregator.commons.Token;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.EncryptedAssertion;
import org.opensaml.saml2.core.Response;

/**
 * SAMLResponseMessage
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class SAMLResponseMessage extends ResponseMessage {

    private Response samlResponse;
    private List<Assertion> assertions = null;
    private String inResponseTo = null;
    private List<Token> tokens = null;

    public SAMLResponseMessage(Response samlResponse) {
        this.samlResponse = samlResponse;
    }

    @Override
    public String getId() {
        return samlResponse.getID();
    }

    @Override
    public String getInResponseTo() {
        if (inResponseTo == null) {
            inResponseTo = SAMLUtil.searchInResponseTo(samlResponse);
        }
        return inResponseTo;
    }

    private List<Assertion> getAssertions() {
        if (assertions == null) {
            assertions = new ArrayList<Assertion>();
            if (samlResponse.getAssertions() != null && !samlResponse.getAssertions().isEmpty()) {
                assertions.addAll(SAMLUtil.getAssertions(samlResponse));
            }
            if (samlResponse.getEncryptedAssertions() != null && !samlResponse.getEncryptedAssertions().isEmpty()) {
                try {
                    assertions.addAll(decryptAssertions(samlResponse.getEncryptedAssertions()));
                } catch (Exception ex) {
                    Logger.getLogger(SAMLResponseMessage.class.getName()).log(Level.SEVERE, null, ex);
                }
            }
        }
        return assertions;
    }

    private List<Assertion> decryptAssertions(List<EncryptedAssertion> encryptedAssertions) throws Exception {
        List<Assertion> decryptedAssertions = new ArrayList<Assertion>();
        Iterator<EncryptedAssertion> it = encryptedAssertions.iterator();
        while (it.hasNext()) {
            EncryptedAssertion encryptedAssertion = it.next();
            decryptedAssertions.add(SAMLUtil.decryptAssertion(encryptedAssertion, API_PKCS12_PATH, API_PKCS12_PASSWORD, API_PKCS12_CERT_ALIAS, API_PRIVATE_KEY_PASSWORD));
        }
        return decryptedAssertions;
    }

    @Override
    public List<Token> getTokens() {
        if (tokens == null) {
            tokens = new ArrayList<Token>();
            Iterator<Assertion> it = getAssertions().iterator();
            while (it.hasNext()) {
                Assertion assertion = it.next();
                if (!assertion.getAuthnStatements().isEmpty()) {
                    try {
                        AuthenticationToken authToken = new AuthenticationToken(assertion);
                        tokens.add(authToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
                if (!assertion.getAttributeStatements().isEmpty()) {
                    try {
                        AttributeToken attributeToken = new AttributeToken(assertion);
                        tokens.add(attributeToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
                if (!assertion.getAuthzDecisionStatements().isEmpty()) {
                    try {
                        AuthorizationToken authorizationToken = new AuthorizationToken(assertion);
                        tokens.add(authorizationToken);
                    } catch (Exception ex) {
                        ex.printStackTrace();
                    }
                }
            }
        }
        return tokens;
    }

    @Override
    public List<Attribute> getAttributes() {
        List<Attribute> attributes = new ArrayList<Attribute>();
        Iterator<Token> it = getTokens().iterator();
        while (it.hasNext()) {
            Token token = it.next();
            if (token instanceof AttributeToken) {
                AttributeToken attributeToken = (AttributeToken) token;
                List<Attribute> attributesAux = attributeToken.getAttributes();
                attributes.addAll(attributesAux);
            }
        }
        return attributes;
    }
    private String API_PKCS12_PATH, API_PKCS12_PASSWORD, API_PKCS12_CERT_ALIAS, API_PRIVATE_KEY_PASSWORD;

    public void setCredentialsInfo(String API_PKCS12_PATH, String API_PKCS12_PASSWORD, String API_PKCS12_CERT_ALIAS, String API_PRIVATE_KEY_PASSWORD) {
        this.API_PKCS12_PATH = API_PKCS12_PATH;
        this.API_PKCS12_PASSWORD = API_PKCS12_PASSWORD;
        this.API_PKCS12_CERT_ALIAS = API_PKCS12_CERT_ALIAS;
        this.API_PRIVATE_KEY_PASSWORD = API_PRIVATE_KEY_PASSWORD;
    }

    @Override
    public String getStatusCode() {
        return samlResponse.getStatus().getStatusCode().getValue();
    }

    @Override
    public String getStatusMessage() {
        return samlResponse.getStatus().getStatusMessage().getMessage();
    }

    @Override
    public String getIssuer() {
        return samlResponse.getIssuer().getValue();
    }

    @Override
    public String getDomain() {
        if (!getAssertions().isEmpty()) {
            if (getAssertions().get(0).getSubject().getNameID() != null) {
                return getAssertions().get(0).getSubject().getNameID().getNameQualifier();
            }
        }
        return null;
    }
}
