/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.identityaggregator.commons.Provider;
import eu.semiramis.identityaggregator.Audit.Level;
import eu.semiramis.policymanagement.query.XACMLAuthorizationQuery;
import eu.semiramis.policymanagement.query.ideas.impl.IDEASXACMLAuthorizationQueryFactory;

import eu.semiramis.policymanagement.query.ideas.impl.IDEASXACMLRequestHelper;
import eu.semiramis.policymanagement.util.XACMLUtils;
import javax.servlet.http.HttpServletRequest;
import org.opensaml.xml.signature.SignableXMLObject;

/**
 * Identity Aggregator - Trust Manager
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class TrustManager {

    protected IdentityAggregator_Config config;
    protected Audit audit;
    static protected final String ACTION_PREFIX = "service/";

    public TrustManager(IdentityAggregator_Config config) {
        this.config = config;
        audit = new Audit(TrustManager.class.getName());
    }

    /*
    public DECISION evaluate(String subject, String action, String resource, String role) throws Exception{
    audit.audit(Level.DEBUG, "Evaluating authorization with IDEAS: [Subject="+subject+", Action="+action+", Resource="+resource+", Role="+role+"]");
    // Create the IDEASAuthZQueryImpl
    IDEASXACMLAuthorizationQueryImpl authZQuery = new IDEASXACMLAuthorizationQueryImpl(
    new SecurityContext(props), props);
    
    // Get the XACML RequestType object and the helper
    IDEASXACMLRequestHelper helper = new IDEASXACMLRequestHelper(XACMLUtils
    .create(org.opensaml.xacml.ctx.RequestType.class, org.opensaml.xacml.ctx.RequestType.DEFAULT_ELEMENT_NAME));
    
    // set attributes
    helper.setInvokingServiceId(subject);
    helper.setFunctionalityId(action);
    helper.setInvokedServiceId(resource);
    
    if(role != null){                        
    helper.setRoleName(role);
    }
    
    // print the request
    audit.audit(Level.DEBUG, "XACMLRequest: "+XACMLUtils.printXMLObject(helper.getRequest()));
    
    // ask for the authorization decision
    org.opensaml.xacml.ctx.ResponseType response = authZQuery.getResponse(helper.getRequest());
    
    // print the decision
    audit.audit(Level.DEBUG, "XACMLResponse: "+XACMLUtils.printXMLObject(response));
    
    return response.getResult().getDecision().getDecision();
    }
     *
     *
     */
    public Decision evaluate(HttpServletRequest request, String subject, String action, String resource, String role) throws Exception {
        // Get the XACML RequestType object and the helper        
        IDEASXACMLRequestHelper helper = new IDEASXACMLRequestHelper(XACMLUtils.create(org.opensaml.xacml.ctx.RequestType.class, org.opensaml.xacml.ctx.RequestType.DEFAULT_ELEMENT_NAME));
        // set attributes
        helper.setInvokingServiceId(subject);
        helper.setFunctionalityId(ACTION_PREFIX + action);
        helper.setInvokedServiceId(resource);

        if (role != null) {
            helper.setRoleName(role);
        }
        // print the request
        audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "XACMLRequest: " + XACMLUtils.printXMLObject(helper.getRequest()));

        //create the IDEASXACMLAuthorizationQueryFactory
        IDEASXACMLAuthorizationQueryFactory authZFactory = new IDEASXACMLAuthorizationQueryFactory();

        //create the XACMLAuthorizationQuery object
        XACMLAuthorizationQuery authZQuery = authZFactory.createXACMLAuthorizationQuery();

        // ask for the authorization decision
        org.opensaml.xacml.ctx.ResponseType response = authZQuery.getResponse(helper.getRequest());

        // print the decision
        audit.audit(Level.DEBUG, request, Audit.AppLogEventId.POLICY_EVALUATION_INFO, "XACMLResponse: " + XACMLUtils.printXMLObject(response));
        return new Decision(response.getResult().getDecision().getDecision());
    }

    public String getCertificatePath(HttpServletRequest request, String issuer) {
        audit.audit(Level.DEBUG, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Recovering [" + issuer + "] certitificate");
        if (!config.Providers.containsKey(issuer)) {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, "No Service Provider info found for Issuer \"" + issuer + "\"");
            return null;
        }
        Provider provider = config.Providers.get(issuer);
        String certPath = provider.getCertPath();
        return certPath;
    }

    public boolean validateSign(HttpServletRequest request, SignableXMLObject samlResponse, String issuer) {
        audit.audit(Level.DEBUG, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Recovering [" + issuer + "] certitificate");
        if (!config.Providers.containsKey(issuer)) {
            audit.audit(Level.ERROR, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, "No Service Provider info found for Issuer \"" + issuer + "\"");
            return false;
        }
        Provider provider = config.Providers.get(issuer);
        if (provider.isCheckSignatures()) {
            audit.audit(Level.DEBUG, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Validating SAMLRequest Signature...");
            try {
                SAMLUtil.validateSign(samlResponse, provider.getCertPath());
                audit.audit(Level.DEBUG, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_INFO, "Signature valid.");
                return true;
            } catch (Exception ex) {
                audit.audit(Level.ERROR, request, Audit.AppLogEventId.SIGNATURE_VALIDATION_ERROR, ex);
                return false;
            }
        }
        return true;
    }
}
