/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.commons.Conditions;
import eu.semiramis.identityaggregator.client.IAClient;
import eu.semiramis.identityaggregator.client.RequestResult;
import eu.semiramis.identityaggregator.client.RequestResult.Status;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.identityaggregator.commons.Token;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.servlet.*;
import javax.servlet.http.*;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;
import org.opensaml.xacml.ctx.DecisionType.DECISION;

/**
 * Identity Aggregator Server Component
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IAServletBasic extends HttpServlet {

    protected IARequestProcessor iaServer;
    protected IAClient iaClient;
    protected AttributeRequestProcessor attributeRequestProcessor;

    @Override
    public void init() {
        Transaction initializationTransaction = new Transaction();
        try {
            iaServer = new IARequestProcessor(getConfigFilePath(), initializationTransaction);
            iaClient = new IAClient(getClientConfigFilePath());
            attributeRequestProcessor = new AttributeRequestProcessor(iaClient, iaServer.getAuthzProvider(), iaServer.getDiscovery(), iaServer.getAttributeTranslator(), iaServer.getID_Translator(), iaServer.getConfig());
        } catch (Exception ex) {
            ex.printStackTrace();
            iaServer.getAudit().audit(Audit.Level.ERROR, initializationTransaction, Audit.AppLogEventId.IA_INITIALIZATION_ERROR, ex);
        }
    }

    protected String getConfigFilePath() {
        return readFile("/WEB-INF/configFilePath.txt").trim();
    }

    protected String getClientConfigFilePath() {
        return readFile("/WEB-INF/clientConfigFilePath.txt").trim();
    }

    private String readFile(String file) {
        ByteArrayInputStream b = null;
        try {
            URL url = this.getServletContext().getResource(file);
            b = (ByteArrayInputStream) url.getContent();
            int length = b.available();
            byte[] buff = new byte[length];
            b.read(buff);
            return new String(buff);
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            try {
                b.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        return null;
    }

    protected void processRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        iaServer.getAudit().initRequest(request);
        try {
            RequestMessage requestMessage = iaServer.processRequest(request);
            if (requestMessage == null) {
                iaServer.showErrorWeb(request, response);
                iaServer.getAudit().audit_UNKNOWN_MESSAGE_RECEIVED(Audit.Level.ERROR, request);
//            } else if (!requestMessage.getStatus().equals(StatusCode.SUCCESS_URI)) {
//                sendErrorBack(request, requestMessage, response, requestMessage.getStatus());
            } else {
                switch (requestMessage.getType()) {
                    case AUTHENTICATION:
                        if (requestMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                            processAuthenticationRequest(requestMessage, request, response);
                        } else {
                            Response samlResponse = sendErrorBack(request, requestMessage, response, requestMessage.getStatusCode(), requestMessage.getStatusMessage());
                            if (samlResponse != null) {
                                iaServer.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                            }
                        }
                        break;
                    case ATTRIBUTE:
                        if (requestMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                            processAttributeRequest(requestMessage, request, response);
                        } else {
                            Response samlResponse = sendErrorBack(request, requestMessage, response, requestMessage.getStatusCode(), requestMessage.getStatusMessage());
                            if (samlResponse != null) {
                                iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                            }
                        }
                        break;
                    case AUTHORIZATION:
                        if (requestMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                            processAuthorizationRequest(requestMessage, request, response);
                        } else {
                            Response samlResponse = sendErrorBack(request, requestMessage, response, requestMessage.getStatusCode(), requestMessage.getStatusMessage());
                            if (samlResponse != null) {
                                iaServer.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                            }
                        }
                        break;
                    default:
                        iaServer.getAudit().audit(Audit.Level.WARN, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, "Not supported request type");
                        sendErrorBack(request, requestMessage, response, StatusCode.REQUEST_UNSUPPORTED_URI, "Unsupported request type.");
                        iaServer.getAudit().audit_UNKNOWN_MESSAGE_RECEIVED(Audit.Level.ERROR, request);
                }
            }
        } finally {
            iaServer.getAudit().clearRequest(request);
        }
    }

    protected Response sendErrorBack(HttpServletRequest request, RequestMessage requestMessage, HttpServletResponse response, String statusCode, String statusMessage) {
        try {
            return iaClient.sendResponse(requestMessage.getId(), null, request, response, statusCode, statusMessage, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            ex.printStackTrace();
            /**
             * @TODO show web error to the end user
             */
            return null;
        }
    }

    /** 
     * Handles the HTTP <code>GET</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Returns a short description of the servlet.
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }

    protected void processAuthenticationRequest(RequestMessage requestMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            iaServer.getAudit().audit_AUTHN_REQ_RECEIVED(Audit.Level.INFO, request, requestMessage.getId(), requestMessage.getIssuer());

            Destination destination = iaServer.getDiscovery().discoverAuthnProviderURL(request, requestMessage);
            if (destination != null) {
                iaClient.setIaId(destination.getProviderId());
                String originIssuer = iaClient.getOriginIssuer(request);

                // Check policies
                Decision decision = iaServer.getAuthzProvider().isAuthenticationAllowed(request, requestMessage.getIssuer(), originIssuer, destination);
                if (!decision.getDecision().equals(DECISION.Permit)) {
                    iaServer.getAudit().audit(Audit.Level.TRACE, request, Audit.AppLogEventId.UNAUTHORIZED_REQUEST, "Unauthorized request");
                    Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.REQUEST_DENIED_URI, "Authentication not permitted by policies.");
                    if (samlResponse != null) {
                        iaServer.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                    }
                    return;
                }

                Conditions conditions = requestMessage.getConditions();
                RequestResult requestResult = iaClient.requestAuthenticationToken(request, response, originIssuer, conditions);
                if (requestResult.getStatus() == Status.REDIRECTED) {
                    RequestLinker.linkRequest(requestResult.getId(), requestMessage);
                    iaServer.getAudit().audit_AUTHN_REQ_SENT(Audit.Level.INFO, request, requestResult.getId(), destination.getProviderId());
                } else if (requestResult.getStatus() == Status.ERROR) {
                    Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error sending authentication request.");
                    if (samlResponse != null) {
                        iaServer.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                    }
                }
            } else {
                iaServer.getDiscovery().showWAYF(request, response, requestMessage);
            }
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing authentication request.");
            if (samlResponse != null) {
                iaServer.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

//    private Map<Destination, SubjectAttributesList> translateProvidersMap(RequestMessage requestMessage, Map<Destination, List<Attribute>> providersMap) {
//        Map<Destination, SubjectAttributesList> providersAttributeMap = new HashMap<Destination, SubjectAttributesList>();
//        Iterator<Destination> it = providersMap.keySet().iterator();
//        while (it.hasNext()) {
//            Destination destination = it.next();
//            List<Attribute> attributes = providersMap.get(destination);
//            String subjectId;
//            if (destination.getDomain() != null) {
//                attributes = translateRequestAttributes(requestMessage.getIssuer(), destination.getDomain(), attributes);
//                subjectId = translateRequestSubjectId(requestMessage.getIssuer(), destination.getDomain(), requestMessage.getSubjectId());
//            }
//            attributes = translateRequestAttributes(requestMessage.getIssuer(), destination.getProviderId(), attributes);
//            subjectId = translateRequestSubjectId(requestMessage.getIssuer(), destination.getProviderId(), requestMessage.getSubjectId());
//
//            SubjectAttributesList subjectAttributeList = new SubjectAttributesList();
//            subjectAttributeList.setAttributes(attributes);
//            subjectAttributeList.setSubjectId(subjectId);
//            providersAttributeMap.put(destination, subjectAttributeList);
//        }
//        return providersAttributeMap;
//    }
    protected void processAttributeRequest(RequestMessage requestMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            iaServer.getAudit().audit_ATTR_REQ_RECEIVED(Audit.Level.INFO, request, requestMessage.getId(), requestMessage.getIssuer());
            if (iaServer.getConfig().AUTO_AUTHENTICATION == true) {
                List<Token> tokens = new ArrayList<Token>();
                tokens.addAll(requestMessage.getTokens());
                List<Token> authnTokens = RequestLinker.getAuthenticationTokens(requestMessage.getId());
                if (authnTokens != null) {
                    tokens.addAll(authnTokens);
                }
                boolean found = false;
                for (Token token : tokens) {
                    if ((token instanceof AuthenticationToken) || (token instanceof AuthorizationToken)) {
                        found = true;
// TODO - Check if valid authentication or authorization tokens are present for the subjectId specified in request? This is currently being checked at AttrP.
                    }
                }
                if (!found) {
                    iaServer.getAudit().audit(Audit.Level.DEBUG, request, Audit.AppLogEventId.REQUEST_PROCESSING_INFO, "Requesting user authentication...");

                    Destination destination = iaServer.getDiscovery().discoverAuthnProviderURL(request, requestMessage);
                    if (destination != null) {
                        iaClient.setIaId(destination.getProviderId());
                        String originIssuer = iaClient.getOriginIssuer(request);
                        // Check policies
                        Decision decision = iaServer.getAuthzProvider().isAuthenticationAllowed(request, requestMessage.getIssuer(), originIssuer, destination);
                        if (!decision.getDecision().equals(DECISION.Permit)) {
                            iaServer.getAudit().audit(Audit.Level.TRACE, request, Audit.AppLogEventId.UNAUTHORIZED_REQUEST, "Unauthorized request");
                            RequestLinker.removeAuthenticationTokens(requestMessage.getId());
                            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.REQUEST_DENIED_URI, "Authentication not permitted by policies.");
                            if (samlResponse != null) {
                                iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                            }
                            return;
                        }

                        RequestResult requestResult = iaClient.requestAuthenticationToken(request, response, originIssuer);
                        if (requestResult.getStatus() == Status.REDIRECTED) {
                            RequestLinker.linkRequest(requestResult.getId(), new SAMLRequestMessage(requestResult.getSAMLRequest()));
                            Map<HttpServletRequest, RequestMessage> attributeRequest = new HashMap<HttpServletRequest, RequestMessage>();
                            attributeRequest.put(request, requestMessage);
                            RequestLinker.linkAuthenticationRequest(requestResult.getId(), attributeRequest);
                            iaServer.getAudit().audit_AUTHN_REQ_SENT(Audit.Level.INFO, request, requestResult.getId(), destination.getProviderId());
                        } else if (requestResult.getStatus() == Status.ERROR) {
                            RequestLinker.removeAuthenticationTokens(requestMessage.getId());
                            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error requesting external attributes.");
                            if (samlResponse != null) {
                                iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                            }
                        }
                    } else {
                        iaServer.getDiscovery().showWAYF(request, response, requestMessage);
                    }
                } else {
                    attributeRequestProcessor.processAttributeRequest(requestMessage, request, response);
                }
            } else {
                attributeRequestProcessor.processAttributeRequest(requestMessage, request, response);
//                iaServer.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, "No subject Id specified in request.");
//                Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.UNKNOWN_PRINCIPAL_URI, "No subject Id specified in request.");
//                if (samlResponse != null) {
//                    iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
//                }
            }
//        List<String> attributeNames = new ArrayList<String>();
//        for (Attribute attribute : requestMessage.getRequestedAttributes()) {
//            attributeNames.add(attribute.getId());
//        }
//        if (!iaServer.getAuthzProvider().isAttributeQueryAllowed(requestMessage.getIssuer(), attributeNames)) {
//            iaServer.getAudit().audit(Audit.Level.TRACE, "Unauthorized request");
//            sendErrorBack(requestMessage, response, StatusCode.REQUEST_DENIED_URI);
//            return;
//        }
//        Map<Destination, List<Attribute>> providersMap = iaServer.getDiscovery().discoverAttrProviderURL(request, requestMessage);
//        //String url = iaServer.getDiscovery().discoverAttrProviderURL(request, requestMessage);
//        if (!providersMap.isEmpty()) {
//            Map<Destination, SubjectAttributesList> providersAttributesMap = translateProvidersMap(requestMessage, providersMap);
//            Destination destination = providersAttributesMap.keySet().iterator().next();
//            List<Attribute> attributes = providersAttributesMap.get(destination).getAttributes();
//            String subjectId = providersAttributesMap.get(destination).getSubjectId();
//            iaClient.setIaId(destination.getProviderId());
//            String originIssuer = iaClient.getOriginIssuer(request);
//            RequestResult requestResult = iaClient.requestAttributeAssertion(request, response, attributes, subjectId, destination.getDomain(), originIssuer, requestMessage.getTokens());
//            if (requestResult.getStatus() == Status.REDIRECTED) {
//                RequestLinker.linkRequest(requestResult.getId(), requestMessage);
//                providersAttributesMap.remove(destination);
//                RequestLinker.linkProvidersMap(requestMessage, providersAttributesMap);
//            } else if (requestResult.getStatus() == Status.ERROR) {
//                sendErrorBack(requestMessage, response, StatusCode.RESPONDER_URI);
//            }
//        } else {
//            iaServer.getDiscovery().showWAYF(response, requestMessage);
//        }
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
            RequestLinker.removeAuthenticationTokens(requestMessage.getId());
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing attribute request.");
            if (samlResponse != null) {
                iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

    protected void processAuthorizationRequest(RequestMessage requestMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            iaServer.getAudit().audit_AUTHZ_REQ_RECEIVED(Audit.Level.INFO, request, requestMessage.getId(), requestMessage.getIssuer());

            Destination destination = iaServer.getDiscovery().discoverAuthzProviderURL(request, requestMessage);
            if (destination != null) {
                iaClient.setIaId(destination.getProviderId());
                String originIssuer = iaClient.getOriginIssuer(request);

                // Check policies
                Decision decision = iaServer.getAuthzProvider().isAuthorizationAllowed(request, requestMessage.getIssuer(), originIssuer, destination);
                if (!decision.getDecision().equals(DECISION.Permit)) {
                    iaServer.getAudit().audit(Audit.Level.TRACE, request, Audit.AppLogEventId.UNAUTHORIZED_REQUEST, "Unauthorized request");
                    Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.REQUEST_DENIED_URI, "Authorization not permitted by policies.");
                    if (samlResponse != null) {
                        iaServer.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                    }
                    return;
                }

                RequestResult requestResult = iaClient.requestAuthorizationToken(request, response, requestMessage.getSubjectId(), destination.getDomain(), requestMessage.getActionId(), requestMessage.getResourceId(), originIssuer, requestMessage.getTokens());
                if (requestResult.getStatus() == Status.REDIRECTED) {
                    RequestLinker.linkRequest(requestResult.getId(), requestMessage);
                    iaServer.getAudit().audit_AUTHZ_REQ_SENT(Audit.Level.INFO, request, requestResult.getId(), destination.getProviderId());
                } else if (requestResult.getStatus() == Status.ERROR) {
                    Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error sending authorization request.");
                    if (samlResponse != null) {
                        iaServer.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                    }
                }
            } else {
                iaServer.getDiscovery().showWAYF(request, response, requestMessage);
            }
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing authorization request.");
            if (samlResponse != null) {
                iaServer.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }
//    protected List<Attribute> translateRequestAttributes(String issuer, String destination, List<Attribute> attributes) {
//        return iaServer.getAttributeTranslator().translateRequestAttributes(issuer, destination, attributes);
//    }
//
//    private String translateRequestSubjectId(String issuer, String destination, String subjectId) {
//        return iaServer.getID_Translator().translateRequestSubjectId(issuer, destination, subjectId);
//    }
}
