/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.SAMLAttribute;
import eu.semiramis.commons.SAMLUtil;
import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.commons.Attribute;
import eu.semiramis.identityaggregator.commons.AttributeToken;
import eu.semiramis.identityaggregator.client.IAClient;
import eu.semiramis.identityaggregator.client.RequestResult;
import eu.semiramis.identityaggregator.client.RequestResult.Status;
import eu.semiramis.identityaggregator.commons.AuthenticationToken;
import eu.semiramis.identityaggregator.commons.AuthorizationToken;
import eu.semiramis.identityaggregator.commons.Token;
import java.io.BufferedReader;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.StringTokenizer;
import javax.mail.Authenticator;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.PasswordAuthentication;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import org.opensaml.saml2.core.Assertion;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;

/**
 * Identity Aggregator Server Component - Response processor
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IAServletBasic_CS extends HttpServlet {

    protected IAClient iaClient;
    protected IAConsumerService iaCS;
    protected AttributeRequestProcessor attributeRequestProcessor;

    /**
     * Processes requests for both HTTP <code>GET</code> and <code>POST</code> methods.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    protected void processRequest(HttpServletRequest request, HttpServletResponse response) throws ServletException, IOException {
        iaCS.getAudit().initRequest(request);
        try {
            ResponseMessage responseMessage = iaCS.processResponse(request);
            String inResponseTo = responseMessage.getInResponseTo();
            RequestMessage requestMessage = RequestLinker.getRequest(inResponseTo);
            switch (requestMessage.getType()) {
                case AUTHENTICATION:
                    processAuthnResponse(requestMessage, responseMessage, request, response);
                    break;
                case ATTRIBUTE:
                    processAttrResponse(requestMessage, responseMessage, request, response);
                    break;
                case AUTHORIZATION:
                    processAuthzResponse(requestMessage, responseMessage, request, response);
                    break;
            }
            RequestLinker.removeLink(inResponseTo);
        } finally {
            iaCS.getAudit().clearRequest(request);
        }
    }

    protected void processAuthnResponse(RequestMessage requestMessage, ResponseMessage responseMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            iaCS.getAudit().audit_AUTHN_RESP_RECEIVED(Audit.Level.INFO, request, responseMessage.getId(), responseMessage.getIssuer(), responseMessage.getInResponseTo());
            if (responseMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                // Search if this authentication has been performed for an attribute request
                Map<HttpServletRequest, RequestMessage> attributeRequest = RequestLinker.getAuthenticationLink(requestMessage.getId());
                if (attributeRequest != null) {
                    RequestLinker.removeAuthenticationLink(requestMessage.getId());
                    Map.Entry<HttpServletRequest, RequestMessage> attrRequest = attributeRequest.entrySet().iterator().next();
                    RequestMessage attrRequestMessage = attrRequest.getValue();
//                    HttpServletRequest attrServletRequest = attrRequest.getKey();
//                    Audit.initRequest(attrServletRequest);
                    Audit.initRequest(request);
                    try {
                        requestMessage = attrRequestMessage;    // Assigned to send the error back to the proper issuer in case of exception
                        List<Token> tokens = iaCS.generateTokens(request, responseMessage.getTokens(), attrRequestMessage.getIssuer());
                        RequestLinker.linkAuthenticationTokens(attrRequestMessage.getId(), tokens);
// TODO--
//                        attrRequestMessage.getTokens().addAll(tokens);
//                        if (attrRequestMessage.getSubjectId() == null || attrRequestMessage.getSubjectId().isEmpty()) {
//                            Iterator<Token> it = tokens.iterator();
//                            if (it.hasNext()) {
//                                String subjectId = it.next().getSubjectId();
//                                attrRequestMessage.changeSubjectId(subjectId);
//                            }
//                        }
                        attributeRequestProcessor.processAttributeRequest(attrRequestMessage, request, response);
                    } finally {
//                        Audit.clearRequest(attrServletRequest);
                        Audit.clearRequest(request);
                    }
                } else {
                    List<Token> tokens = iaCS.generateTokens(request, responseMessage.getTokens(), requestMessage.getIssuer());
                    Response samlResponse = iaClient.sendResponse(requestMessage.getId(), tokens, request, response, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
                    iaCS.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.INFO, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());

                    // Audit Authentication token release
                    String subjectId = null;
                    for (Token token : tokens) {
                        if (token instanceof AuthenticationToken) {
                            subjectId = ((AuthenticationToken) token).getSubjectId();
                        }
                    }
                    if (subjectId != null) {
                        iaCS.getAudit().audit_AUTHENTICATION(Audit.Level.INFO, request, requestMessage.getOriginIssuer(), subjectId);
                    }
                }

            } else {
                Response samlResponse = sendErrorBack(request, requestMessage, response, responseMessage.getStatusCode(), responseMessage.getStatusMessage());
                if (samlResponse != null) {
                    iaCS.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                }
            }
        } catch (Exception ex) {
            iaCS.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing authentication response.");
            if (samlResponse != null) {
                iaCS.getAudit().audit_AUTHN_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

    protected List<Attribute> translateResponseAttributes(HttpServletRequest request, String destination, String issuer, List<Attribute> attributes, String subjectId) {
        return iaCS.getAttributeTranslator().translateResponseAttributes(request, destination, issuer, attributes, subjectId);
    }

    protected String translateResponseSubjectId(HttpServletRequest request, String issuer, String destination, String subjectId) {
        return iaCS.getID_Translator().translateResponseSubjectId(request, destination, issuer, subjectId);
    }

    protected void processAttrResponse(RequestMessage requestMessage, ResponseMessage responseMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            iaCS.getAudit().audit_ATTR_RESP_RECEIVED(Audit.Level.INFO, request, responseMessage.getId(), responseMessage.getIssuer(), responseMessage.getInResponseTo());
            if (responseMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                List<Attribute> recoveredAttributes = responseMessage.getAttributes();
                recoveredAttributes = translateResponseAttributes(request, responseMessage.getIssuer(), requestMessage.getIssuer(), recoveredAttributes, requestMessage.getSubjectId());
                if (responseMessage.getDomain() != null) {
                    recoveredAttributes = translateResponseAttributes(request, responseMessage.getDomain(), requestMessage.getIssuer(), recoveredAttributes, requestMessage.getSubjectId());
                }

                List<Token> tokens = new ArrayList<Token>();
                if (!responseMessage.getTokens().isEmpty()) {
                    List<SAMLAttribute> attributes = new ArrayList<SAMLAttribute>();
                    Iterator<Attribute> it = recoveredAttributes.iterator();
                    while (it.hasNext()) {
                        Attribute attribute = it.next();
                        attributes.add(attribute.toSAMLAttribute());
                    }
                    String subjectId = translateResponseSubjectId(request, responseMessage.getIssuer(), requestMessage.getIssuer(), responseMessage.getTokens().get(0).getSubjectId());
                    if (responseMessage.getDomain() != null) {
                        subjectId = translateResponseSubjectId(request, responseMessage.getDomain(), requestMessage.getIssuer(), subjectId);
                    }
                    Assertion assertion = SAMLUtil.createAttrAssertion(subjectId, attributes);
                    AttributeToken attributeToken = new AttributeToken(assertion);
                    tokens.add(attributeToken);
                }
                for (Token token : responseMessage.getTokens()) {
                    if (!(token instanceof AttributeToken)) {
                        tokens.add(token);
                    }
                }

                RequestLinker.LinkResponseTokens(requestMessage, tokens);

                /****/
                /* Check if it should collect more attributes */
                Map<Destination, SubjectAttributesList> providersAttributesMap = RequestLinker.getProvidersMap(requestMessage);
                if (!providersAttributesMap.isEmpty()) {
                    Destination destination = providersAttributesMap.keySet().iterator().next();
                    List<Attribute> attributes = providersAttributesMap.get(destination).getAttributes();
                    String subjectId = providersAttributesMap.get(destination).getSubjectId();
                    iaClient.setIaId(destination.getProviderId());
// TODO - Use transaction from the original attribute request to enable audit linkability, instead of a new Transaction().
                    RequestResult requestResult = iaClient.requestAttributeAssertion((String) null, response, attributes, subjectId, destination.getDomain(), requestMessage.getOriginIssuer(), requestMessage.getAttrEncrX509Cert(), requestMessage.getTokens(), new Transaction());
                    if (requestResult.getStatus() == Status.REDIRECTED) {
                        RequestLinker.linkRequest(requestResult.getId(), requestMessage);
                        providersAttributesMap.remove(destination);
                        //RequestLinker.linkProvidersMap(requestMessage,providersAttributesMap);
                        iaCS.getAudit().audit_ATTR_REQ_SENT(Audit.Level.INFO, request, requestResult.getId(), destination.getProviderId());
                    } else if (requestResult.getStatus() == Status.ERROR) {
                        RequestLinker.removeResponseTokens(requestMessage);
                        Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error sending attribute request.");
                        if (samlResponse != null) {
                            iaCS.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                        }
                    }
                    return;
                }
                /*****/
                List<Token> responseTokens = RequestLinker.getResponseTokens(requestMessage);
                RequestLinker.removeResponseTokens(requestMessage);
                Response samlResponse = iaClient.sendResponse(requestMessage.getId(), responseTokens, request, response, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
                iaCS.getAudit().audit_ATTR_RESP_SENT(Audit.Level.INFO, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());

                /* Audit authorization tokens and attributes released in response */
                List<String> attrNames = new ArrayList<String>();
                for (Token responseToken : responseTokens) {
                    if (responseToken instanceof AttributeToken) {
                        List<Attribute> responseAttributes = ((AttributeToken) responseToken).getAttributes();
                        for (Attribute responseAttribute : responseAttributes) {
                            attrNames.add(responseAttribute.getId());
                        }
                    } else if (responseToken instanceof AuthorizationToken) {
                        AuthorizationToken authorizationToken = (AuthorizationToken) responseToken;
                        iaCS.getAudit().audit_AUTHZ_TOKEN_RELEASE(Audit.Level.INFO, request, authorizationToken.getAudienceURI(), decodeValues(authorizationToken.getResource()), authorizationToken.getSubjectId(), authorizationToken.getNotOnOrAfter());
                    }
                }
                if (!attrNames.isEmpty()) {
                    iaCS.getAudit().audit_ATTR_RELEASE(Audit.Level.INFO, request, requestMessage.getOriginIssuer(), attrNames, requestMessage.getSubjectId());

                    /* User notification according to policies */

                    Object notificationNeeded = request.getSession(true).getAttribute("NotificationNeeded");
                    if ((notificationNeeded != null) && ((Boolean) notificationNeeded)) {
                        String notificationEmailAddress = iaCS.getDiscovery().getNotificationEmailAddress(request, requestMessage.getSubjectId());
                        sendUserNotification(notificationEmailAddress, requestMessage.getSubjectId(), attributeRequestProcessor.getOriginIssuer(requestMessage), attrNames);
                    }
                }
            } else {
                RequestLinker.removeResponseTokens(requestMessage);
                Response samlResponse = sendErrorBack(request, requestMessage, response, responseMessage.getStatusCode(), responseMessage.getStatusMessage());
                if (samlResponse != null) {
                    iaCS.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                }
            }
        } catch (Exception ex) {
            RequestLinker.removeResponseTokens(requestMessage);
            iaCS.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_PROCESSING_ERROR, ex);
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing attribute response.");
            if (samlResponse != null) {
                iaCS.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

    protected void processAuthzResponse(RequestMessage requestMessage, ResponseMessage responseMessage, HttpServletRequest request, HttpServletResponse response) {
        try {
            if (responseMessage.getStatusCode().equals(StatusCode.SUCCESS_URI)) {
                iaCS.getAudit().audit_AUTHZ_RESP_RECEIVED(Audit.Level.INFO, request, responseMessage.getId(), responseMessage.getIssuer(), responseMessage.getInResponseTo());
                iaClient.sendResponse(requestMessage.getId(), responseMessage.getTokens(), request, response, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());

                /* Audit authorization token released in response */
                for (Token responseToken : RequestLinker.getResponseTokens(requestMessage)) {
                    if (responseToken instanceof AuthorizationToken) {
                        AuthorizationToken authorizationToken = (AuthorizationToken) responseToken;
                        iaCS.getAudit().audit_AUTHZ_TOKEN_RELEASE(Audit.Level.INFO, request, authorizationToken.getAudienceURI(), decodeValues(authorizationToken.getResource()), authorizationToken.getSubjectId(), authorizationToken.getNotOnOrAfter());
                    }
                }
            } else {
                Response samlResponse = sendErrorBack(request, requestMessage, response, responseMessage.getStatusCode(), responseMessage.getStatusMessage());
                if (samlResponse != null) {
                    iaCS.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                }
            }
        } catch (Exception ex) {
            iaCS.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            Response samlResponse = sendErrorBack(request, requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing authorization response.");
            if (samlResponse != null) {
                iaCS.getAudit().audit_AUTHZ_RESP_SENT(Audit.Level.ERROR, request, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        }
    }

    protected Response sendErrorBack(HttpServletRequest request, RequestMessage requestMessage, HttpServletResponse response, String statusCode, String statusMessage) {
        try {
            response.reset();
            return iaClient.sendResponse(requestMessage.getId(), null, request, response, statusCode, statusMessage, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
        } catch (Exception ex) {
            iaCS.getAudit().audit(Audit.Level.ERROR, request, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            ex.printStackTrace();
            return null;
        }
    }

    protected String getConfigFilePath() {
        return readFile("/WEB-INF/configFilePath.txt").trim();
    }

    protected String getClientConfigFilePath() {
        return readFile("/WEB-INF/clientConfigFilePath.txt").trim();
    }

    private String readFile(String file) {
        ByteArrayInputStream b = null;
        try {
            URL url = this.getServletContext().getResource(file);
            b = (ByteArrayInputStream) url.getContent();
            int length = b.available();
            byte[] buff = new byte[length];
            b.read(buff);
            return new String(buff);
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            try {
                b.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        return null;
    }

    @Override
    public void init() {
        Transaction initializationTransaction = new Transaction();
        try {
            iaCS = new IAConsumerService(getConfigFilePath(), initializationTransaction);
            iaClient = new IAClient(getClientConfigFilePath());
            attributeRequestProcessor = new AttributeRequestProcessor(iaClient, iaCS.getAuthzProvider(), iaCS.getDiscovery(), iaCS.getAttributeTranslator(), iaCS.getID_Translator(), iaCS.getConfig());
        } catch (Exception ex) {
            ex.printStackTrace();
            iaCS.getAudit().audit(Audit.Level.ERROR, initializationTransaction, Audit.AppLogEventId.IA_CS_INITIALIZATION_ERROR, ex);
        }
    }

    /** 
     * Handles the HTTP <code>GET</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Returns a short description of the servlet.
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }

    private void sendUserNotification(String notificationEmailAddress, String userId, String attributeRequestor, List<String> attributeNames) throws IOException, MessagingException {
        FileInputStream fstream = new FileInputStream(iaCS.getConfig().NOTIFICATION_MAIL_TEMPLATE_PATH);
        BufferedReader in = new BufferedReader(new InputStreamReader(new DataInputStream(fstream)));
        StringWriter msg = new StringWriter();
        PrintWriter out = new PrintWriter(msg);

        StringWriter attributeList = new StringWriter();
        PrintWriter attrList = new PrintWriter(attributeList);

        Iterator<String> attributeNamesIterator = attributeNames.iterator();
        while (attributeNamesIterator.hasNext()) {
            attrList.print("    - " + attributeNamesIterator.next());
            if (attributeNamesIterator.hasNext()) {
                attrList.println();
            }
        }
        attrList.close();

        Date date = new Date();

        String strLine;
        while ((strLine = in.readLine()) != null) {
            strLine = strLine.replace("#USER_ID#", userId);
            strLine = strLine.replace("#ATTRIBUTE_REQUESTOR#", attributeRequestor);
            strLine = strLine.replace("#DATE#", date.toString());
            strLine = strLine.replace("#ATTRIBUTE_LIST#", attributeList.toString());
            out.println(strLine);
        }
        in.close();
        out.close();

        sendMail(notificationEmailAddress, iaCS.getConfig().NOTIFICATION_MAIL_SUBJECT, msg.toString());
    }

    public void sendMail(String address, String subject, String text) throws MessagingException {
        Properties properties = new Properties();
        properties.setProperty("mail.smtp.host", iaCS.getConfig().MAIL_SMTP_HOST);
        if (iaCS.getConfig().MAIL_SMTP_PORT != null) {
            properties.setProperty("mail.smtp.port", iaCS.getConfig().MAIL_SMTP_PORT);
        }
        if (iaCS.getConfig().MAIL_USE_SSL) {
            properties.setProperty("mail.smtp.ssl.enable", "true");
        }

        Session session = null;
        if ((iaCS.getConfig().MAIL_AUTHENTICATION_USER != null) || iaCS.getConfig().MAIL_AUTHENTICATION_PASSWORD != null) {
            properties.setProperty("mail.smtp.auth", "true");
            session = Session.getDefaultInstance(properties, new Authenticator() {

                @Override
                protected PasswordAuthentication getPasswordAuthentication() {
                    String user = iaCS.getConfig().MAIL_AUTHENTICATION_USER;
                    String password = iaCS.getConfig().MAIL_AUTHENTICATION_PASSWORD;
                    if (user == null) {
                        user = iaCS.getConfig().MAIL_FROM_ADDRESS;
                    }
                    if (password == null) {
                        password = "";
                    }
                    return new PasswordAuthentication(user, password);
                }
            });
        } else {
            session = Session.getDefaultInstance(properties);
        }

        Message message = new MimeMessage(session);
        message.addRecipient(Message.RecipientType.TO, new InternetAddress(address));
        message.setSubject(subject);
        message.setText(text);
        message.setFrom(new InternetAddress(iaCS.getConfig().MAIL_FROM_ADDRESS));
        Transport.send(message);
    }
    /*
     * Authorization Token resource field parsing
     */
    private String FIELD_SEPARATOR = ";";
    private String ESCAPE_CHAR = "\\";

    private List<String> decodeValues(String string) {
        List<String> values = new ArrayList<String>();
        if (string != null) {
            String escapedValue = "";
            StringTokenizer tokenizer = new StringTokenizer(string, FIELD_SEPARATOR);
            while (tokenizer.hasMoreTokens()) {
                String token = tokenizer.nextToken();
                if (token.endsWith(ESCAPE_CHAR)) {
                    escapedValue = token.substring(0, token.length() - ESCAPE_CHAR.length()) + FIELD_SEPARATOR;
                } else {
                    values.add(escapedValue + token);
                    escapedValue = "";
                }
            }
            if (!escapedValue.isEmpty()) {
                values.add(escapedValue);
            }
        }
        return values;
    }
}
