/*
 *    Copyright (C) 2012, 2013 Universidad de Murcia
 *
 *    Authors:
 *        Ginés Dólera Tormo <ginesdt@um.es>
 *        Juan M. Marín Pérez <juanmanuel@um.es>
 *        Jorge Bernal Bernabé <jorgebernal@um.es>
 *        Gregorio Martínez Pérez <gregorio@um.es>
 *        Antonio F. Skarmeta Gómez <skarmeta@um.es>
 *
 *    This file is part of SEMIRAMIS Identity Aggregator (SEMIRAMIS-IA).
 *
 *    SEMIRAMIS-IA is free software: you can redistribute it and/or modify
 *    it under the terms of the GNU Lesser General Public License as published by
 *    the Free Software Foundation, either version 3 of the License, or
 *    (at your option) any later version.
 *
 *    SEMIRAMIS-IA is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU Lesser General Public License for more details.
 *
 *    You should have received a copy of the GNU Lesser General Public License
 *    along with SEMIRAMIS-IA. If not, see <http://www.gnu.org/licenses/>.
 *
 */
package eu.semiramis.identityaggregator;

import eu.semiramis.commons.audit.Transaction;
import eu.semiramis.identityaggregator.client.IAClient;
import eu.semiramis.identityaggregator.commons.Attribute;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.http.HttpSession;
import org.opensaml.saml2.core.Response;
import org.opensaml.saml2.core.StatusCode;

/**
 * IAServletConsent
 *
 * @author Ginés Dólera Tormo
 * @author Juan M. Marín Pérez
 * @author Jorge Bernal Bernabé
 * @author Gregorio Martínez Pérez
 * @author Antonio F. Skarmeta Gómez
 */
public class IAServletConsent extends HttpServlet {

    protected IARequestProcessor iaServer;
    protected IAClient iaClient;
    protected AttributeRequestProcessor attributeRequestProcessor;

    @Override
    public void init() {
        Transaction initializationTransaction = new Transaction();
        try {
            iaServer = new IARequestProcessor(getConfigFilePath(), initializationTransaction);
            iaClient = new IAClient(getClientConfigFilePath());
            attributeRequestProcessor = new AttributeRequestProcessor(iaClient, iaServer.getAuthzProvider(), iaServer.getDiscovery(), iaServer.getAttributeTranslator(), iaServer.getID_Translator(), iaServer.getConfig());
        } catch (Exception ex) {
            ex.printStackTrace();
            iaServer.getAudit().audit(Audit.Level.ERROR, initializationTransaction, Audit.AppLogEventId.IA_INITIALIZATION_ERROR, ex);
        }
    }

    /** 
     * Processes requests for both HTTP <code>GET</code> and <code>POST</code> methods.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    protected void processRequest(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {

        iaServer.getAudit().initRequest(request);
        Transaction transaction = iaServer.getAudit().getRequestTransaction(request);

        HttpSession session = request.getSession();
        RequestMessage requestMessage = (RequestMessage) session.getAttribute("RequestMessage");
        session.removeAttribute("RequestMessage");
        String requestURL = (String) session.getAttribute("RequestURL");
        session.removeAttribute("RequestURL");
        Map<Destination, List<Attribute>> providersMap = (Map<Destination, List<Attribute>>) session.getAttribute("ProvidersMap");
        session.removeAttribute("ProvidersMap");
        String[] released = request.getParameterValues("released");
        try {
            if (!request.getParameter("approvalDecission").equals("authorize") || (released == null) || (released.length == 0)) {
                iaServer.getAudit().audit(Audit.Level.TRACE, transaction, Audit.AppLogEventId.UNAUTHORIZED_REQUEST, "Attribute request not authorized by user consent.");
                Response samlResponse = sendErrorBack(requestMessage, response, StatusCode.REQUEST_DENIED_URI, "Attribute release not authorized by user.", transaction);
                if (samlResponse != null) {
                    iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, transaction, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
                }
                return;
            }
            Map<Destination, List<Attribute>> approvedProvidersMap = new HashMap<Destination, List<Attribute>>();
            List<String> approvedAttributes = new ArrayList<String>();
            approvedAttributes.addAll(Arrays.asList(released));
            for (Map.Entry<Destination, List<Attribute>> entry : providersMap.entrySet()) {
                Destination destination = entry.getKey();
                for (Attribute attribute : entry.getValue()) {
                    if (approvedAttributes.contains(attribute.getId())) {
                        List<Attribute> destinationAttributes = approvedProvidersMap.get(destination);
                        if (destinationAttributes == null) {
                            destinationAttributes = new ArrayList<Attribute>();
                            approvedProvidersMap.put(destination, destinationAttributes);
                        }
                        if (!destinationAttributes.contains(attribute)) {
                            destinationAttributes.add(attribute);
                        }
                    }

                }
            }
            attributeRequestProcessor.requestAttributes(requestMessage, response, approvedProvidersMap, requestURL, transaction);
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, transaction, Audit.AppLogEventId.REQUEST_PROCESSING_ERROR, ex);
            Response samlResponse = sendErrorBack(requestMessage, response, StatusCode.RESPONDER_URI, "Internal server error processing user consent.", transaction);
            if (samlResponse != null) {
                iaServer.getAudit().audit_ATTR_RESP_SENT(Audit.Level.ERROR, transaction, samlResponse.getID(), requestMessage.getIssuer(), samlResponse.getInResponseTo());
            }
        } finally {
            Audit.clearRequest(request);
        }
    }

    protected Response sendErrorBack(RequestMessage requestMessage, HttpServletResponse response, String statusCode, String statusMessage, Transaction transaction) {
        try {
            return iaClient.sendResponse(requestMessage.getId(), null, response, statusCode, statusMessage, requestMessage.getIssuer(), requestMessage.getConsumerServiceURL());
        } catch (Exception ex) {
            iaServer.getAudit().audit(Audit.Level.ERROR, transaction, Audit.AppLogEventId.RESPONSE_GENERATION_ERROR, ex);
            ex.printStackTrace();
            /**
             * @TODO show web error to the end user
             */
            return null;
        }
    }

    // <editor-fold defaultstate="collapsed" desc="HttpServlet methods. Click on the + sign on the left to edit the code.">
    /** 
     * Handles the HTTP <code>GET</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Handles the HTTP <code>POST</code> method.
     * @param request servlet request
     * @param response servlet response
     * @throws ServletException if a servlet-specific error occurs
     * @throws IOException if an I/O error occurs
     */
    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
            throws ServletException, IOException {
        processRequest(request, response);
    }

    /** 
     * Returns a short description of the servlet.
     * @return a String containing servlet description
     */
    @Override
    public String getServletInfo() {
        return "Short description";
    }// </editor-fold>

    protected String getConfigFilePath() {
        return readFile("/WEB-INF/configFilePath.txt").trim();
    }

    protected String getClientConfigFilePath() {
        return readFile("/WEB-INF/clientConfigFilePath.txt").trim();
    }

    private String readFile(String file) {
        ByteArrayInputStream b = null;
        try {
            URL url = this.getServletContext().getResource(file);
            b = (ByteArrayInputStream) url.getContent();
            int length = b.available();
            byte[] buff = new byte[length];
            b.read(buff);
            return new String(buff);
        } catch (IOException ex) {
            ex.printStackTrace();
        } finally {
            try {
                b.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }
        }
        return null;
    }
}
